#include "gsmhacking.h"

struct blended_position **alloc_blended_positions(unsigned short n) {
  return (struct blended_position **) alloc_lots(n, "blended_position", sizeof(struct blended_position *), (void *(*)()) alloc_blended_position);
}

struct blended_position *alloc_blended_position() {
  return (struct blended_position *) alloc_one("blended_position", sizeof(struct blended_position));
}

void cleanup_blended_positions(unsigned short num_blended_positions, struct blended_position **blended_positions) {
  unsigned short i;
  for (i = 0; i < num_blended_positions; i++) {
    free_structs("blended_position_short", blended_positions[i]->num_positions, (void **) blended_positions[i]->positions);
    free_structs("blended_position_float", blended_positions[i]->num_positions, (void **) blended_positions[i]->weights);
  }
  free_structs("blended_positions", num_blended_positions, (void *) blended_positions);
}


/* find out how many blended positions there are and what their offsets are */
int find_blended_positions(struct gsm *gsm) {
  struct blended_position *blended_position;
  unsigned int pos;
  unsigned short i;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think blended positions start at 0x%04x\n", pos);

  /* how many blended_positions are there? */
  if (read_short(gsm->fd, &gsm->num_blended_positions)) {
    ERROR("can't read number of blended positions");
    return 2;
  }
  /* length of blended position stream */
  if (read_int(gsm->fd, &gsm->blended_position_stream_length)) {
    ERROR("can't read blended position stream length");
    return 3;
  }

  /* if there's no stream, might as well return */
  if (! gsm->num_blended_positions && ! gsm->blended_position_stream_length) {
    gsm->blended_positions = 0;
    return 0;
  }

  if (gsm->number_blended_positions != gsm->num_blended_positions) {
    ERROR("number of blended positions mismatched (%d v %u)\n", gsm->number_blended_positions, gsm->num_blended_positions); 
    return 4;
  }

  gsm->blended_positions = alloc_blended_positions(gsm->num_blended_positions);
  if (! gsm->blended_positions) return 4;

  for (i = 0; i < gsm->num_blended_positions; i++) {
    blended_position = gsm->blended_positions[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    blended_position->offset = pos;
    if (parse_blended_position(gsm, i, blended_position, gsm->num_blended_positions)) {
      ERROR("expected blended position stream but got null\n");
      return 6;
    }
  }

  if (get_file_position(gsm->fd) != gsm->blended_positions[0]->offset + gsm->blended_position_stream_length) {
    ERROR("blended position stream mismatch (%d > %d + %u)\n", get_file_position(gsm->fd), gsm->blended_positions[0]->offset, gsm->blended_position_stream_length);
    return 0;
  }

  return 0;
}

int parse_blended_position(struct gsm *gsm, unsigned short num, struct blended_position *blended_position, unsigned short total) {
  unsigned short i;
  float sum;

  if (read_short(gsm->fd, &blended_position->num_positions)) {
    ERROR("read failed parsing blended position number\n");
    return 1;
  }

  blended_position->positions = (unsigned short **) alloc_lots(blended_position->num_positions, "short",  sizeof(unsigned short *), (void *(*)()) alloc_short);
  if (! blended_position->positions) {
    ERROR("out of memory for shorts list");
    return 2;
  }
  blended_position->weights = (float **) alloc_lots(blended_position->num_positions, "float",  sizeof(float *), (void *(*)()) alloc_float);
  if (! blended_position->weights) {
    ERROR("out of memory for floats list");
    return 3;
  }

  sum = 0;
  for (i = 0; i < blended_position->num_positions; i++) {
    if (read_short(gsm->fd, blended_position->positions[i])) {
      ERROR("read failed parsing blended position entry position\n");
      return 4;
    }
    if (read_float(gsm->fd, blended_position->weights[i])) {
      ERROR("read failed parsing blended position entry weight\n");
      return 5;
    }
    TRACE1("offset 0x%04x, position %u entry %u: vertex %u, weight %0.4f\n", blended_position->offset, num, i, *blended_position->positions[i], *blended_position->weights[i]);
    if (*blended_position->positions[i] > gsm->num_positions) {
      WARN("oops, blended position entry %u has position outside range num_positions\n", num);
      WARN("please tell furrycat about %s\n", gsm->basename);
    }
    sum += *blended_position->weights[i];
  }
  TRACE("blended position %u at 0x%04x: positions: %u with sum %0.4f\n", num, blended_position->offset, blended_position->num_positions, sum);
  if ((1 - sum < -0.0001 || 1 - sum > 0.0001) && (sum < -0.0001 || sum > 0.0001)) {
    WARN("blended position %u weights are not normalised (sum %f)!\n", num, sum);
  }

  return 0;
}

int write_blended_positions(struct gsm *gsm) {
  unsigned short i, j;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing blended positions: %u (length %u)\n", gsm->num_blended_positions, gsm->blended_position_stream_length);

  /* write actual stream */
  if (write_short(gsm->out_fd, &gsm->num_blended_positions)) {
    ERROR("can't write number of blended positions");
    return 4;
  }
  if (write_int(gsm->out_fd, &gsm->blended_position_stream_length)) {
    ERROR("can't write blended position stream length");
    return 5;
  }

  for (i = 0; i < gsm->num_blended_positions; i++) {
    if (write_short(gsm->out_fd, &gsm->blended_positions[i]->num_positions)) {
      ERROR("can't write blended position number of lists");
      return 6;
    }
    for (j = 0; j < gsm->blended_positions[i]->num_positions; j++) {
      if (write_short(gsm->out_fd, gsm->blended_positions[i]->positions[j])) {
        ERROR("can't write blended position entry position %u", j);
        return 7;
      }
      if (write_float(gsm->out_fd, gsm->blended_positions[i]->weights[j])) {
        ERROR("can't write blended position entry weight %u", j);
        return 7;
      }
    }
  }

  return 0;
}

int describe_blended_positions(struct gsm *gsm) {
  unsigned short i, j;
  struct blended_position *blended_position;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Blended position stream length: %d in %u positions\n", gsm->blended_position_stream_length, gsm->num_blended_positions);
  if (gsm_verbose) {
    SAY("  First blended position index: %u\n", gsm->first_blended_position);
    if (gsm_verbose > 1) {
      SAY("\n");
      for (i = 0; i < gsm->num_blended_positions; i++) {
        blended_position = gsm->blended_positions[i];
        SAY("  Blended position %u: blends %u positions\n", i, blended_position->num_positions);
        for (j = 0; j < blended_position->num_positions; j++) {
          SAY("    Position %04u with weight %0.4f\n", *blended_position->positions[j], *blended_position->weights[j]);
        }
        SAY("\n");
      }
    }
  }
  return 0;
}
