#ifndef BONE_H
#define BONE_H
#include "data.h"
#include "maths.h"
#include <stdlib.h>

#define BONE_FIRST_LINK 0xffff
#define BONE_COORDS_LENGTH 30

GSM_API struct bone **alloc_bones(unsigned short);
GSM_API struct bone *alloc_bone();
GSM_API void cleanup_bones(unsigned short, struct bone **, unsigned short);
GSM_API int find_bones();
GSM_API int parse_bone(struct gsm *, unsigned short, struct bone *, unsigned short);
GSM_API struct bone **make_swt_bones();
GSM_API int make_swt_bone(unsigned short, struct bone *);
GSM_API struct bone **make_kid_bones();
GSM_API int make_kid_bone(unsigned short, struct bone *);
GSM_API struct bone **make_tac_bones();
GSM_API int make_tac_bone(unsigned short, struct bone *);
GSM_API struct bone **make_bf1_bones();
GSM_API int make_bf1_bone(unsigned short, struct bone *);
GSM_API int write_bones(struct gsm *, struct skeleton *);
GSM_API char *get_bone_by_id(struct gsm *, unsigned short);
GSM_API char *get_bone_link(struct gsm *, unsigned short);
GSM_API int get_bone_by_name(struct gsm *, char *);
GSM_API int find_real_bone_position(struct gsm *, unsigned short);
GSM_API vector_t *bone_local_to_world(struct bone *, struct bone *);
GSM_API int calculate_bone_matrices(struct bone *);
GSM_API int describe_bones(struct gsm *);

struct bone {
  unsigned short number;
  off_t offset;
  unsigned short name_length;
  char *name;
  unsigned short links_to;
  struct bone *parent;
  vector_t position; /* XXX: pointers to elements of matrix? */
  quaternion_t q; /* default pose frame */
  quaternion_t world_q; /* kill me */
  unsigned char animated; /* for animation; not exported if 0 */
  quaternion_t **orientations; /* kill me */
  quaternion_t **world_orientations; /* kill me */
  vector_t **positions; /* kill me */
  /* rewrite */
  matrix_t local;
  matrix_t world;
  matrix_t **locals;
  matrix_t **worlds;
};

/* damn gonna have to do this properly */
#ifndef __cplusplus
/* data stream is (links_to, x, y, z, q->w, q->x ,q->y, q->z) */
static unsigned char swt_bone_coords[SWT_BIPED_NUM_BONES][BONE_COORDS_LENGTH] = {
  "\xff\xff\xbc\x93\xd3\x3b\xaf\x4f\x15\x42\xf3\x89\x58\xbf\x39\xff\xff\xbe\x99\xff\xff\xbe\x10\x02\x00\x3f\x0b\xfd\xff\xbe"
 ,"\x00\x00\x83\x07\x34\x40\x50\xf8\x82\xbb\x18\x70\xba\x36\x93\x4a\x0c\x36\xbe\xc5\x36\x35\x74\x2e\xd3\x3c\x38\xea\x7f\x3f"
 ,"\x01\x00\x25\xe3\xa0\x40\xa8\x46\xa2\xbb\x5f\x93\x41\x36\xe7\x11\x91\x34\x8c\x46\x0e\x33\xf5\xf7\x8e\xbc\x05\xf6\x7f\x3f"
 ,"\x02\x00\xc1\x14\xc7\x40\xc7\xa0\xd3\xbb\xe1\x50\x71\x36\xb6\x82\x98\xb4\x00\x00\x78\xaa\x00\x58\x04\x2e\x00\x00\x80\x3f"
 ,"\x03\x00\x33\xc3\x01\x41\x8c\xfa\x4a\xbb\xcc\x28\x9d\x36\x87\x02\xd3\x32\xe0\x4e\x0d\x35\x70\x87\x50\xbe\xba\xa2\x7a\x3f"
 ,"\x04\x00\xf1\x9b\x87\x40\x0c\x32\x8e\xb5\xba\x9b\x27\x36\x50\xd9\x9d\xba\x79\x6c\x89\xbb\x16\xa6\xe8\x3d\x25\x57\x7e\x3f"
 ,"\x04\x00\x22\x85\x5b\xb5\xc9\xd2\x5c\x3b\x9c\x70\xcf\x3f\xb7\xa8\x23\x3f\xf8\x6f\x08\xbe\x76\xcc\x3d\x3f\x29\x3a\x1e\xbe"
 ,"\x06\x00\x67\x21\xcc\x40\x1f\x79\x6a\xb6\xcb\xf7\xd7\x36\x97\x7e\xbc\xbc\xe6\x61\x16\x3d\x16\x71\xe2\x3c\x63\xa9\x7f\x3f"
 ,"\x07\x00\x08\x01\x42\x41\x84\xb9\xe1\xb6\xb3\x8e\xf5\x36\x2e\x49\x56\xb3\xc1\xf5\x3c\x32\x22\x6a\x47\x3e\x4c\x19\x7b\x3f"
 ,"\x08\x00\x8f\xfc\x26\x41\x0d\x50\xce\xb6\x6b\xee\x08\x37\xc4\x1e\x7f\x3f\x4a\x80\xa9\x3d\x42\x33\xad\xb8\xc2\xa9\x61\x3b"
 ,"\x09\x00\xfc\x6a\x36\x3f\x18\x99\x0d\x3f\xf7\x32\xb6\xbf\xf9\xa9\xdc\xbe\x3c\xc3\x95\xbe\x63\xe6\xe7\xbb\xeb\x85\x5a\x3f"
 ,"\x0a\x00\x20\x42\xcf\x3f\x5a\xcc\xe7\xb4\xf4\x3d\x82\x36\x01\x9a\xf2\x32\x72\x0b\x56\xb2\x06\xf2\x85\xbd\xaf\x73\x7f\x3f"
 ,"\x0b\x00\xfe\x30\x93\x3f\x52\x8b\xda\xb6\x07\x7b\x3e\xb6\xc2\xeb\xb0\x32\x7b\x3b\x92\xb2\xff\x0d\x7a\xbd\xc3\x85\x7f\x3f"
 ,"\x09\x00\x5e\x0d\x81\x40\x45\x26\x3d\xbe\x3a\xd3\xa0\xbf\x8f\xe1\xf4\xbc\x14\x96\x15\xbd\xf9\x44\x63\xbd\xe6\x51\x7f\x3f"
 ,"\x0d\x00\x4c\x85\xc3\x3f\xf1\x1a\x9d\xb4\x3c\xa3\x1e\x36\x61\xaa\xf2\x31\x5e\xc5\x05\xb2\xa0\x96\xf9\xbd\x81\x17\x7e\x3f"
 ,"\x0e\x00\xe7\xfc\x95\x3f\x60\x85\x2b\xb6\x00\x4e\x67\xb5\xc3\x6d\x9d\xb0\x39\x32\x81\xb0\xd6\xf8\x8e\xbc\x05\xf6\x7f\x3f"
 ,"\x09\x00\x75\xc8\x85\x40\xf0\x9d\x9e\xbe\xc3\x18\x32\xbe\xe7\x30\xba\x3a\x79\x00\x11\xbc\x07\x49\xb2\xbd\x8f\x04\x7f\x3f"
 ,"\x10\x00\x8d\xd0\xeb\x3f\x04\x54\x98\xb6\x9c\xc0\x9f\xb5\x5f\xce\x42\xb3\x57\xe5\xf8\xb1\x8e\xf5\xde\xbd\x7a\x7a\x7e\x3f"
 ,"\x11\x00\x24\x09\x99\x3f\x13\x63\x59\xb6\x52\x02\xed\xb5\x29\xa8\x4d\xb2\xd7\x2b\xa6\x2f\x61\x2c\xfa\xbc\x6f\xe1\x7f\x3f"
 ,"\x09\x00\xd6\xb1\x78\x40\x16\x29\x14\xbe\xb5\x66\x46\x3f\xdb\x4a\xe9\x3d\xd6\x82\x78\x3d\x8e\x4d\xe7\xbd\x06\x35\x7c\x3f"
 ,"\x13\x00\x19\x45\xe7\x3f\xb1\x2e\xc3\xb4\x60\x34\xea\xb2\xcf\xf1\x9f\xb0\xee\x36\xe3\x31\x41\x13\xd6\xbd\xfc\x98\x7e\x3f"
 ,"\x14\x00\x49\xdb\xa1\x3f\xf3\xab\x4e\xb6\x7c\x98\x2a\xb6\xbb\x41\xd8\xb1\xa8\xad\x8a\xb0\xd2\xf7\x8e\x3c\x05\xf6\x7f\x3f"
 ,"\x09\x00\xe2\x21\x62\x40\x25\x2d\x9e\x3e\x0c\x3f\xc6\x3f\xc3\xd5\x61\x3e\x6b\xbf\x1b\x3e\xcf\x5f\x05\xbe\xc6\x60\x74\x3f"
 ,"\x16\x00\x18\xa2\xbe\x3f\x58\x85\xaf\xb5\x96\x05\x56\x35\x4e\xa9\xe6\x30\xd8\x15\xf1\x32\x3c\x97\xa9\xbd\xeb\x1e\x7f\x3f"
 ,"\x17\x00\xa9\xfa\x79\x3f\x8e\x81\x27\x36\x71\xb5\x36\xb6\xd4\xc2\xfb\x31\x9a\xab\xd7\xb0\xf2\xf8\x0e\x3c\x81\xfd\x7f\x3f"
 ,"\x04\x00\x09\x00\x6d\xb5\xb5\x6b\x5d\x3b\x9a\x70\xcf\xbf\xaf\xa8\x23\xbf\x68\x6f\x08\x3e\x7c\xcc\x3d\x3f\xa3\x3a\x1e\xbe"
 ,"\x19\x00\x6c\x21\xcc\x40\x87\x1f\x04\xb6\x68\x58\x48\x36\x96\x7e\xbc\x3c\xe8\x61\x16\xbd\x41\x71\xe2\x3c\x63\xa9\x7f\x3f"
 ,"\x1a\x00\x0a\x01\x42\x41\x2e\x77\x8e\xb6\xd8\x98\xb7\x36\x28\xac\x28\xb3\x22\xa5\x1d\xb3\x23\x6a\x47\x3e\x4c\x19\x7b\x3f"
 ,"\x1b\x00\x90\xfc\x26\x41\x0a\xd9\x7b\xb6\x1d\xdd\xa0\x36\xc4\x1e\x7f\x3f\x4a\x80\xa9\x3d\x3f\x3a\xad\x38\x5a\xaa\x61\xbb"
 ,"\x1c\x00\x8d\x62\x33\x3f\x7d\xf3\x10\x3f\x18\x26\xb8\x3f\xdf\x62\xdc\x3e\xb7\xda\x95\x3e\xec\xc9\x62\xbc\x64\x8e\x5a\x3f"
 ,"\x1d\x00\x06\x42\xcf\x3f\x13\xdd\x55\x35\x32\x7e\x3f\xb5\xcf\x93\x44\xb2\x86\x96\xd4\x2f\x01\xf2\x85\xbd\xaf\x73\x7f\x3f"
 ,"\x1e\x00\xfe\x30\x93\x3f\x03\xfc\x5d\xb5\x63\x13\x60\xb5\x3e\xd0\xa2\xb0\x9a\xfc\x1f\x33\x02\x0e\x7a\xbd\xc3\x85\x7f\x3f"
 ,"\x1c\x00\x3c\xf1\x80\x40\x15\x9f\x2f\x3b\xd9\x6a\xa1\x3f\x83\xe1\xf4\x3c\x17\x96\x15\x3d\xfc\x44\x63\xbd\xe6\x51\x7f\x3f"
 ,"\x20\x00\x5c\x85\xc3\x3f\x60\xc2\x3c\x34\x27\xbc\xd7\xb6\xd3\xa9\x5d\xb3\x58\x69\x18\x33\xa0\x96\xf9\xbd\x81\x17\x7e\x3f"
 ,"\x21\x00\xdd\xfc\x95\x3f\x9a\xca\xc3\xb5\x32\x32\x99\x35\xe2\xe2\x1e\xb2\x1f\x7a\xc3\x31\xdc\xf8\x8e\xbc\x05\xf6\x7f\x3f"
 ,"\x1c\x00\xaf\x40\x84\x40\x12\x97\xff\xbd\x0b\x00\x5b\x3e\x0d\x3d\x7a\xbb\x8a\x76\x11\x3c\x7d\x2d\x68\xbd\x90\x93\x7f\x3f"
 ,"\x23\x00\x6d\xd0\xeb\x3f\x96\x05\xc6\xb5\x84\x60\x24\xb5\x29\x66\x39\x32\xe3\xa0\x25\x33\x8d\xf5\xde\xbd\x7a\x7a\x7e\x3f"
 ,"\x24\x00\x20\x09\x99\x3f\x42\x7e\xf8\xb5\x85\x22\x12\xb5\x33\x69\xae\x31\x86\xd9\x2a\x32\x6e\x2c\xfa\xbc\x6f\xe1\x7f\x3f"
 ,"\x1c\x00\xf7\xab\x76\x40\x60\xed\xab\x3d\xae\x62\x42\xbf\xe3\x4a\xe9\xbd\xd8\x82\x78\xbd\x97\x4d\xe7\xbd\x06\x35\x7c\x3f"
 ,"\x26\x00\x1a\x45\xe7\x3f\xb1\x9a\xd0\xb5\xb8\x6d\x7e\x36\x06\x52\xef\xb2\xb6\x8f\xa2\x32\x41\x13\xd6\xbd\xfc\x98\x7e\x3f"
 ,"\x27\x00\x33\xdb\xa1\x3f\x3d\x27\x80\xb5\x68\xdb\xe6\xb5\x36\xce\x71\x32\x65\xbf\x80\xb1\xd5\xf7\x8e\x3c\x05\xf6\x7f\x3f"
 ,"\x1c\x00\xaf\x2d\x63\x40\x1f\x80\xc1\x3e\xc2\x8d\xc8\xbf\xc7\xd5\x61\xbe\x6e\xbf\x1b\xbe\xd7\x5f\x05\xbe\xc6\x60\x74\x3f"
 ,"\x29\x00\x20\xa2\xbe\x3f\x93\x38\x30\xb6\x03\x10\x1b\x35\xc7\x1f\xf7\xb2\x27\x93\x45\xb2\x40\x97\xa9\xbd\xeb\x1e\x7f\x3f"
 ,"\x2a\x00\xad\xfa\x79\x3f\x64\xc9\xbd\x34\x17\xd7\x50\x35\xaa\x1c\x5e\x31\xa8\x9a\xb6\x31\x18\xf9\x0e\x3c\x81\xfd\x7f\x3f"
 ,"\x00\x00\x3a\x41\xec\xb5\x40\x70\xd1\x36\xd3\xe0\x66\x40\x24\xc3\x06\xbd\x06\xda\x7f\x3f\xf8\xcd\x96\xb9\xdd\xe3\x0e\x3c"
 ,"\x2c\x00\x71\xf7\x98\x41\x88\xe5\x59\xb7\x63\xda\x41\x37\x15\x9a\xd7\xb1\x1f\xc5\xf1\x30\x3d\x70\xd3\x3c\x2b\xea\x7f\x3f"
 ,"\x2d\x00\xc4\x6e\x72\x41\x6a\x81\x24\xb7\x9c\x43\x15\x37\x76\x98\x81\x38\x6c\xf9\x0e\x3c\x51\x60\x70\xbd\x8d\x8c\x7f\x3f"
 ,"\x2e\x00\x97\xf9\x8e\x40\x2b\xda\x58\x40\x17\x12\x06\x36\xa0\x2f\x08\x34\x63\x22\x5b\xb4\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
 ,"\x00\x00\x96\xbb\x6f\x36\x39\x01\x4b\xb6\xd2\xe0\x66\xc0\xe7\xc0\x06\xbd\x07\xda\x7f\x3f\x03\x1a\x96\x39\x85\xe9\x0e\xbc"
 ,"\x30\x00\x71\xf7\x98\x41\x5f\x28\x5b\xb7\xc3\x71\x39\x37\x55\xa4\xd0\xb1\x68\x7a\x8e\x30\x04\x7a\xd3\x3c\x29\xea\x7f\x3f"
 ,"\x31\x00\xc1\x6e\x72\x41\x81\x2a\x29\xb7\xbf\xb1\x13\x37\xb5\x56\x81\xb8\xb8\xf9\x0e\xbc\xf3\x62\x70\xbd\x8a\x8c\x7f\x3f"
 ,"\x32\x00\x97\xf9\x8e\x40\x29\xda\x58\x40\x02\x2c\x08\x36\xa0\x2f\x08\x34\x63\x22\x5b\xb4\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
};

static unsigned char kid_bone_coords[KID_BIPED_NUM_BONES][BONE_COORDS_LENGTH] = {
  "\xff\xff\x7b\xdf\xed\x3b\x00\x1e\xbe\x41\x54\x5c\x58\xbf\x39\xff\xff\xbe\x99\xff\xff\xbe\x10\x02\x00\x3f\x0b\xfd\xff\xbe"
 ,"\x00\x00\x0b\x09\x34\x40\x88\x13\x10\xbb\x05\x50\xba\x36\xcb\x13\x0a\x36\x8a\x55\x34\x35\x73\x2e\xd3\x3c\x38\xea\x7f\x3f"
 ,"\x01\x00\xec\xf9\x30\x40\x04\x80\x32\xbb\xa6\x49\xd6\x35\xc6\xee\xb0\x31\x39\x50\x61\x33\xbb\xf7\x8e\xbc\x05\xf6\x7f\x3f"
 ,"\x02\x00\x3b\xfd\x5a\x40\x95\xf9\x7e\xbb\x14\x02\x05\x36\x40\x97\x86\x31\x00\x00\x50\x2a\xc0\x54\x00\x30\x00\x00\x80\x3f"
 ,"\x03\x00\x52\x54\x9c\x40\x05\x18\x00\xbb\x7c\xe7\x3d\x36\x7f\xa3\xc7\x32\x58\x2e\xba\x34\x53\xb7\xf0\xbd\xbc\x39\x7e\x3f"
 ,"\x04\x00\x89\x70\x21\x40\x66\xb1\x33\xb5\x80\x03\xcb\x35\x39\xae\x5a\xba\xef\x54\x8c\xbb\x6a\xb3\xd9\x3c\x3b\xe8\x7f\x3f"
 ,"\x05\x00\x49\x32\xcc\x40\xdd\x1d\x57\xc0\x6a\x1d\x46\x40\x4c\xaa\x01\xbf\x61\x6d\x4a\x3f\x68\xda\xed\xbd\x24\xad\xa5\x3e"
 ,"\x06\x00\x38\x06\x4f\x40\x04\x8a\x3a\xba\xa0\xfb\xf0\xba\x44\x77\xb3\xbe\x3e\x15\x53\xbe\x31\x2a\xf7\xbe\xb8\x8f\x46\x3f"
 ,"\x07\x00\x57\x8f\x47\x40\xf0\xc0\xa7\xbb\xbc\xb4\x96\x35\x5d\xb2\x0d\x33\xaa\x39\x41\x35\x6e\xf4\x83\xbe\x19\x5a\x77\x3f"
 ,"\x05\x00\x70\x18\xce\x40\xd7\x22\x59\xc0\xee\xa0\x3f\xc0\xa7\xa9\x01\xbf\x0c\x6d\x4a\x3f\x94\xea\xed\x3d\x5b\xaf\xa5\xbe"
 ,"\x09\x00\xdd\x07\x4f\x40\x79\x12\xd2\xba\x8f\xbf\x84\x3a\x1b\x13\x49\x3e\x50\x56\x92\x3d\x5a\x68\xbe\xbe\x33\x8a\x67\x3f"
 ,"\x0a\x00\x8c\xa3\x47\x40\xb3\x13\x89\xbb\x30\x71\x5a\x35\x72\x7e\x81\xb1\xf3\xaf\x91\x35\xd1\x65\xc3\xbe\xc1\x9f\x6c\x3f"
 ,"\x04\x00\xf0\x1f\xdd\x3e\xf3\xd4\xdb\x3e\x62\xce\xa0\x3f\x51\x02\x26\x3f\xc5\xb9\x9d\xbd\x2a\x86\x40\x3f\xe4\xe9\xb6\xbd"
 ,"\x0c\x00\x39\xb3\x90\x40\x18\xd3\x2e\xb6\x99\xcd\x86\x36\xcf\x7e\xbc\xbc\xe4\x61\x16\x3d\x10\x71\xe2\x3c\x63\xa9\x7f\x3f"
 ,"\x0d\x00\x09\x9c\xea\x40\x7c\xb3\x8b\xb6\x07\x6f\x31\x36\x23\x0d\x54\xb3\x74\xd4\x3d\x32\x22\x6a\x47\x3e\x4c\x19\x7b\x3f"
 ,"\x0e\x00\xf1\xdb\xd2\x40\xca\x6b\x87\xb6\x6c\xb8\xad\x36\xc4\x1e\x7f\x3f\x4e\x80\xa9\x3d\x76\x33\xbd\xb8\x05\xd3\x61\x3b"
 ,"\x0f\x00\xd1\x7c\x5f\x3f\x13\xc9\x01\x3f\x7e\x4f\x4a\xbf\x4d\xe4\x38\xbf\x1b\xd4\x6a\xbe\x76\x26\xd6\x3d\xe2\xe2\x24\x3f"
 ,"\x10\x00\x00\x3b\x93\x3f\xb2\x48\xd4\xb3\x81\x92\x57\xb5\xdb\x23\xb4\xb2\x9b\x8d\x9d\x31\x03\xf2\x85\xbd\xaf\x73\x7f\x3f"
 ,"\x0f\x00\x00\x73\x27\x40\xbe\xa2\x2a\xbc\x80\x29\x06\xbf\xa0\x14\x83\x39\xbe\xd3\x0e\x3c\xca\xfc\x8e\xbc\x86\xf3\x7f\x3f"
 ,"\x12\x00\x03\x9b\xcf\x3f\x78\xc2\x6c\x33\xb1\x97\x45\xb6\x61\x5e\x47\xb3\x1c\xc7\x11\x32\x9e\x96\xf9\xbd\x81\x17\x7e\x3f"
 ,"\x0f\x00\x2a\xb9\x27\x40\x05\x60\xda\xba\xb9\xf1\x01\x3f\x6f\x24\x70\x3a\xf9\xc4\x20\x3d\xd0\x83\x56\xbc\xd9\xc7\x7f\x3f"
 ,"\x14\x00\x3b\x9e\xb4\x3f\x75\x2a\x80\xb5\x13\x4e\x11\xb6\xd2\x9c\x1d\xb3\x51\xb5\x8c\xb2\x89\xf5\xde\xbd\x7a\x7a\x7e\x3f"
 ,"\x04\x00\x92\x1f\xdd\x3e\xf8\xd5\xdb\x3e\x58\xce\xa0\xbf\x4c\x02\x26\xbf\xa3\xb8\x9d\x3d\x2e\x86\x40\x3f\xd4\xea\xb6\xbd"
 ,"\x16\x00\x3a\xb3\x90\x40\x6e\x00\xcf\xb5\x82\x93\xb1\xb5\xce\x7e\xbc\x3c\xe6\x61\x16\xbd\x3a\x71\xe2\x3c\x63\xa9\x7f\x3f"
 ,"\x17\x00\x09\x9c\xea\x40\x45\xc2\x39\xb6\x52\x77\xaf\x36\x12\x1a\x2a\xb3\x29\xa9\x14\xb3\x23\x6a\x47\x3e\x4c\x19\x7b\x3f"
 ,"\x18\x00\xf6\xdb\xd2\x40\x8b\x28\x0a\xb6\x0e\x6f\xc8\x36\xc4\x1e\x7f\x3f\x4f\x80\xa9\x3d\xef\x38\xbd\x38\xa1\xd3\x61\xbb"
 ,"\x19\x00\x73\xf5\x72\x3f\x9e\x29\x74\x3e\xe3\x0a\x4d\x3f\xae\x27\x39\x3f\xee\xd0\x8f\x3e\x27\xbc\x9f\x3d\xf6\x42\x20\x3f"
 ,"\x1a\x00\xed\x3a\x93\x3f\x45\x27\x3f\xb5\x1a\xaf\x88\x35\xe4\x02\xc6\x32\xb7\xee\xc9\xb2\x03\xf2\x85\xbd\xaf\x73\x7f\x3f"
 ,"\x19\x00\xe5\x72\x27\x40\x09\x75\x28\xbc\x68\x29\x06\x3f\xc4\x15\x83\xb9\xc4\xd3\x0e\xbc\xd4\xfc\x8e\xbc\x86\xf3\x7f\x3f"
 ,"\x1c\x00\x19\x9b\xcf\x3f\xc8\x56\x98\xb5\xb4\x49\x1d\x36\x81\x8b\x70\xb1\x72\x96\x8b\x32\xa2\x96\xf9\xbd\x81\x17\x7e\x3f"
 ,"\x19\x00\x31\xb9\x27\x40\x89\x64\xda\xba\x15\xf2\x01\xbf\x0e\x25\x70\xba\xf8\xc4\x20\xbd\xd6\x83\x56\xbc\xd9\xc7\x7f\x3f"
 ,"\x1e\x00\x33\x9e\xb4\x3f\xbc\x54\x10\xb5\xab\xde\x02\x36\xb1\xbe\x05\xb2\xf9\xfd\x11\x33\x8f\xf5\xde\xbd\x7a\x7a\x7e\x3f"
 ,"\x00\x00\x3f\xf1\x99\xb5\xa0\xc4\xac\x36\x29\x24\x4e\x40\x66\x74\x08\xbd\x20\xd9\x7f\x3f\x05\xea\x94\xb9\x57\xe4\x0e\x3c"
 ,"\x20\x00\xf5\xd5\x36\x41\x27\x38\x01\xb7\x7a\x10\xdf\x36\xc0\x9d\xd6\xb1\x68\x6d\x04\x30\x23\x7b\xd4\x3c\xf4\xe9\x7f\x3f"
 ,"\x21\x00\x69\x96\x06\x41\x93\x0a\xc0\xb6\xf9\xf3\xa2\x36\x12\x75\x6f\x38\x6f\xf9\x0e\x3c\x70\x96\x72\xbd\x76\x8a\x7f\x3f"
 ,"\x22\x00\x3b\xea\x75\x40\x7a\xcd\x1f\x40\x26\x39\xa8\x35\xe9\x2f\x08\x34\xa6\x22\x5b\xb4\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
 ,"\x00\x00\x56\xd8\x9d\x36\x5d\x4a\x50\xb6\x29\x24\x4e\xc0\x2a\x71\x08\xbd\x22\xd9\x7f\x3f\x1e\x37\x94\x39\x08\xea\x0e\xbc"
 ,"\x24\x00\xf6\xd5\x36\x41\x23\x1e\x01\xb7\x2e\x5c\xdd\x36\x72\x07\xd4\xb1\x2c\xde\xd2\xaf\xde\x8a\xd4\x3c\xf0\xe9\x7f\x3f"
 ,"\x25\x00\x6a\x96\x06\x41\x6b\x0c\xc1\xb6\x41\x28\xa9\x36\x5d\xc4\x6e\xb8\xf9\xf9\x0e\xbc\x0e\x9b\x72\xbd\x71\x8a\x7f\x3f"
 ,"\x26\x00\x39\xea\x75\x40\x7b\xcd\x1f\x40\x27\xbf\xab\x35\xe9\x2f\x08\x34\xa6\x22\x5b\xb4\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
};

static unsigned char tac_bone_coords[TAC_BIPED_NUM_BONES][BONE_COORDS_LENGTH] = {
  "\xff\xff\xbc\x93\xd3\x3b\xaf\x4f\x15\x42\xf3\x89\x58\xbf\x39\xff\xff\xbe\x99\xff\xff\xbe\x10\x02\x00\x3f\x0b\xfd\xff\xbe"
 ,"\x00\x00\x83\x07\x34\x40\x50\xf8\x82\xbb\x18\x70\xba\x36\x93\x4a\x0c\x36\xbe\xc5\x36\x35\x74\x2e\xd3\x3c\x38\xea\x7f\x3f"
 ,"\x01\x00\x2d\xe3\xa0\x40\x02\x45\xa2\xbb\x91\x8f\x41\x36\xe5\x11\x91\x34\x60\x46\x0e\x33\xbc\xf7\x8e\xbc\x05\xf6\x7f\x3f"
 ,"\x02\x00\xc1\x14\xc7\x40\x78\xa1\xd3\xbb\x6d\x4d\x71\x36\xb8\x82\x98\xb4\x00\x00\x50\xaa\x00\x6c\xef\xaf\x00\x00\x80\x3f"
 ,"\x03\x00\x33\xc3\x01\x41\x8a\xf9\x4a\xbb\x42\x2f\x9d\x36\x4d\x02\xd3\x32\xd9\x4e\x0d\x35\x69\x87\x50\xbe\xba\xa2\x7a\x3f"
 ,"\x04\x00\xfb\x9b\x87\x40\x47\xd0\x14\xb6\x0c\x9d\x27\x36\x49\xd9\x9d\xba\x7a\x6c\x89\xbb\xfc\xa5\xe8\x3d\x25\x57\x7e\x3f"
 ,"\x05\x00\x44\x72\x2e\x40\xdd\xc5\x2e\xc0\x5a\xd2\xd3\xb6\x99\xb4\xa4\xb3\x0f\x89\xe0\xb5\x0f\xb5\x09\x3f\x9e\xce\x57\x3f"
 ,"\x06\x00\x1f\xdf\x9a\x40\x75\x59\x7c\xbb\xd0\xad\xad\x35\x31\x1e\x86\x31\x7e\xb2\x08\x34\x0a\x07\x2e\xbd\xd3\xc4\x7f\x3f"
 ,"\x07\x00\x27\xdf\x9a\x40\xdb\x70\x7b\xbb\xd9\x33\xdb\x35\xda\xa8\xa2\x31\x19\x7d\x08\x34\x12\x07\x2e\xbd\xd3\xc4\x7f\x3f"
 ,"\x04\x00\x86\xc8\x66\x35\xe6\xcd\x5c\x3b\x9c\x70\xcf\x3f\xb7\xa8\x23\x3f\xf4\x6f\x08\xbe\x76\xcc\x3d\x3f\x25\x3a\x1e\xbe"
 ,"\x09\x00\x67\x21\xcc\x40\xb0\xfe\x67\xb6\xb4\x6b\xd3\x36\xcf\x7e\xbc\xbc\xe5\x61\x16\x3d\x10\x71\xe2\x3c\x63\xa9\x7f\x3f"
 ,"\x0a\x00\x07\x01\x42\x41\x4a\xea\xe6\xb6\xc2\xf5\xed\x36\x2e\x49\x56\xb3\xc1\xf5\x3c\x32\x22\x6a\x47\x3e\x4c\x19\x7b\x3f"
 ,"\x0b\x00\x90\xfc\x26\x41\x73\xe9\xb8\xb6\x7a\x76\x07\x37\xc4\x1e\x7f\x3f\x4a\x80\xa9\x3d\x42\x33\xad\xb8\xc2\xa9\x61\x3b"
 ,"\x0c\x00\xb2\x58\xea\x3f\xce\xbe\x53\x3f\xd4\xe0\xb4\xbf\x13\x04\x37\xbf\x67\xa5\x80\xbe\x1b\x66\xfe\x3d\x7e\xfd\x23\x3f"
 ,"\x0d\x00\x90\xe6\xce\x3f\x2e\xdc\x47\xb6\xb5\x6f\x84\xb5\x66\xab\xef\xb2\x74\x95\xb6\x32\x01\xf2\x85\xbd\xaf\x73\x7f\x3f"
 ,"\x0c\x00\xd5\xbb\x60\x40\x13\xdd\x3a\xbc\xcd\x8e\x66\xbf\xa3\xf6\xab\x3a\x82\x5c\x56\xbd\xc4\x96\x8e\xbc\x31\x9c\x7f\x3f"
 ,"\x0f\x00\x75\xb1\x36\x40\xb0\xe8\x1c\x35\x2d\xed\x06\xb5\x52\x7c\x37\xb3\xbf\x4c\xab\xb1\xa2\x96\xf9\xbd\x81\x17\x7e\x3f"
 ,"\x0c\x00\x37\x6d\x62\x40\xed\xb5\xc0\xba\x62\x6e\x22\x3f\x8b\x4d\x6f\x3a\x3c\xc6\x20\x3d\xf3\x91\x56\xbc\xd8\xc7\x7f\x3f"
 ,"\x11\x00\xa4\xf1\x1e\x40\x94\x57\x71\xb5\x29\xb6\x80\xb6\xd3\xec\x1e\xb3\x62\x05\x41\xb2\x8b\xf5\xde\xbd\x7a\x7a\x7e\x3f"
 ,"\x04\x00\xe1\x4d\x55\x35\xd3\x66\x5d\x3b\x9a\x70\xcf\xbf\xaf\xa8\x23\xbf\x64\x6f\x08\x3e\x7c\xcc\x3d\x3f\x9f\x3a\x1e\xbe"
 ,"\x13\x00\x6b\x21\xcc\x40\x3e\xa5\x09\xb6\xe4\x7a\x51\x36\xcf\x7e\xbc\x3c\xe7\x61\x16\xbd\x3a\x71\xe2\x3c\x63\xa9\x7f\x3f"
 ,"\x14\x00\x08\x01\x42\x41\x46\xa8\x93\xb6\xa2\x2f\xbf\x36\x28\xac\x28\xb3\x22\xa5\x1d\xb3\x23\x6a\x47\x3e\x4c\x19\x7b\x3f"
 ,"\x15\x00\x91\xfc\x26\x41\x38\x09\x51\xb6\xc7\xcc\xa3\x36\xc4\x1e\x7f\x3f\x4a\x80\xa9\x3d\x3f\x3a\xad\x38\x5a\xaa\x61\xbb"
 ,"\x16\x00\x03\xab\xf2\x3f\x23\x90\xf5\x3e\x56\x3b\xb6\x3f\x86\xe0\x36\x3f\x31\x0c\x9b\x3e\x1f\xe2\xc6\x3d\x50\x93\x1f\x3f"
 ,"\x17\x00\x9e\xe6\xce\x3f\x65\x45\x56\xb6\x2f\x11\xa0\xb5\x25\x4a\x0c\x32\x3c\xfa\x52\xb2\x02\xf2\x85\xbd\xaf\x73\x7f\x3f"
 ,"\x16\x00\xc3\xbb\x60\x40\x7d\xd9\x3a\xbc\xca\x8e\x66\x3f\xab\xf6\xab\xba\x82\x5c\x56\x3d\xc9\x96\x8e\xbc\x31\x9c\x7f\x3f"
 ,"\x19\x00\x64\xb1\x36\x40\x65\x17\x12\x35\x9c\x7e\x02\xb5\x96\xb6\x24\xb2\x3b\x1c\x04\x33\xa3\x96\xf9\xbd\x81\x17\x7e\x3f"
 ,"\x16\x00\x33\x6d\x62\x40\xc3\xa4\xc0\xba\x79\x6e\x22\xbf\x38\x4e\x6f\xba\x3d\xc6\x20\xbd\xfa\x91\x56\xbc\xd8\xc7\x7f\x3f"
 ,"\x1b\x00\x98\xf1\x1e\x40\xd6\x47\x44\xb5\x31\x74\x20\x36\x92\xb6\xe7\xb1\x6e\x17\x02\x33\x8f\xf5\xde\xbd\x7a\x7a\x7e\x3f"
 ,"\x00\x00\x3a\x41\xec\xb5\x40\x70\xd1\x36\xd3\xe0\x66\x40\x24\xc3\x06\xbd\x06\xda\x7f\x3f\xf8\xcd\x96\xb9\xdd\xe3\x0e\x3c"
 ,"\x1d\x00\x71\xf7\x98\x41\x88\xe5\x59\xb7\x63\xda\x41\x37\x15\x9a\xd7\xb1\x1f\xc5\xf1\x30\x3d\x70\xd3\x3c\x2b\xea\x7f\x3f"
 ,"\x1e\x00\xc4\x6e\x72\x41\x6a\x81\x24\xb7\x9c\x43\x15\x37\x76\x98\x81\x38\x6c\xf9\x0e\x3c\x51\x60\x70\xbd\x8d\x8c\x7f\x3f"
 ,"\x1f\x00\x97\xf9\x8e\x40\x2b\xda\x58\x40\x17\x12\x06\x36\xa0\x2f\x08\x34\x63\x22\x5b\xb4\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
 ,"\x00\x00\x96\xbb\x6f\x36\x39\x01\x4b\xb6\xd2\xe0\x66\xc0\xe7\xc0\x06\xbd\x07\xda\x7f\x3f\x03\x1a\x96\x39\x85\xe9\x0e\xbc"
 ,"\x21\x00\x71\xf7\x98\x41\x5f\x28\x5b\xb7\xc3\x71\x39\x37\x55\xa4\xd0\xb1\x68\x7a\x8e\x30\x04\x7a\xd3\x3c\x29\xea\x7f\x3f"
 ,"\x22\x00\xc1\x6e\x72\x41\x81\x2a\x29\xb7\xbf\xb1\x13\x37\xb5\x56\x81\xb8\xb8\xf9\x0e\xbc\xf3\x62\x70\xbd\x8a\x8c\x7f\x3f"
 ,"\x23\x00\x97\xf9\x8e\x40\x29\xda\x58\x40\x02\x2c\x08\x36\xa0\x2f\x08\x34\x63\x22\x5b\xb4\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
};

static unsigned char bf1_bone_coords[BF1_BIPED_NUM_BONES][BONE_COORDS_LENGTH] = {
  "\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x97\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x3f"
 ,"\x00\x00\x30\xd7\x29\x40\xbf\xde\x59\xbb\x2a\xca\x77\x36\xf0\xec\x0a\x36\x15\x46\x36\x35\x73\x2e\xd3\x3c\x38\xea\x7f\x3f"
 ,"\x01\x00\x06\xdc\x85\x40\x89\x43\x86\xbb\x73\xed\x45\xb2\xd7\xe5\x41\x34\x27\x26\x1b\x33\xbb\xf7\x8e\xbc\x05\xf6\x7f\x3f"
 ,"\x02\x00\x74\xbf\xa4\x40\xf4\x3e\xa0\xbb\xb2\x75\x01\xb2\xc0\x9f\x47\xb4\x27\x96\x3b\xa8\x00\x30\x4b\x2b\x00\x00\x80\x3f"
 ,"\x03\x00\xeb\x8e\xc4\x40\xcd\x5d\x36\xbb\xdc\xfa\x3d\xb1\x06\xa2\xf6\x2f\xdf\x36\x00\x35\xb6\x99\x24\xbe\xcb\xab\x7c\x3f"
 ,"\x04\x00\x32\xce\x6b\x40\x3b\xae\x19\x35\x4d\x3d\x2e\x2f\xa4\xae\x85\xba\xdf\x02\x8b\xbb\xc5\xac\x8f\x3d\xe8\x5d\x7f\x3f"
 ,"\x04\x00\xb9\x24\xea\xb5\xef\x20\x40\x3b\x8d\xd5\x92\x3f\xe9\x78\x22\x3f\x46\x37\xd4\xbd\x9d\x7c\x41\x3f\xc1\xb8\xfc\xbd"
 ,"\x06\x00\x20\xed\xa8\x40\x13\xfe\x54\xb3\x50\x19\x37\xb5\x86\x52\xc3\xbc\x8b\x87\x04\x3d\x87\x95\xdc\x3c\x43\xb3\x7f\x3f"
 ,"\x07\x00\xbd\xde\x1b\x41\x2e\x35\x77\xb4\x15\x95\xc7\xb5\x00\x18\x3c\x30\xf8\xf1\x0b\x32\xee\x0d\x2d\x3e\x5a\x51\x7c\x3f"
 ,"\x08\x00\x0c\x0a\x21\x41\x8e\x6f\xd5\x34\x55\x34\x1b\x35\xff\xae\x7f\x3f\xf5\x37\x20\x3d\x79\x55\xf8\x3c\xc9\xe2\x97\x3b"
 ,"\x09\x00\x9a\xf0\x5c\x3f\x27\x20\xe8\x3e\xa6\x47\x8d\xbf\x29\xf7\x2e\xbf\xbb\x66\xb0\xbe\xb4\xc2\x2a\x3e\x14\x20\x1f\x3f"
 ,"\x0a\x00\xf0\x37\xab\x3f\x80\x03\xb9\xb5\x61\x7b\x9a\xb5\x26\x22\xd6\xb2\x8b\xd6\x7d\xb2\x30\x00\x3f\xbe\xde\x81\x7b\x3f"
 ,"\x09\x00\xe3\xf9\x39\x40\xd2\x7d\x2d\xbc\x9f\xcb\x3e\xbf\x06\xa3\xcd\x3b\xb3\xc3\x85\xbd\xbb\x02\x0c\xbd\x64\x4c\x7f\x3f"
 ,"\x0c\x00\x25\xdd\x2d\x40\x1f\x24\x18\x35\xb4\xda\x3a\x35\xe2\xda\xcc\x30\x6f\x3f\x6f\x30\xad\xf2\x0e\xbd\x14\xd8\x7f\x3f"
 ,"\x09\x00\x8a\x60\x3b\x40\xcc\x23\xc5\xba\x54\x6b\x06\x3f\xd6\x22\x95\x3b\xcb\x46\x56\xbc\x81\x9d\xf7\xbc\xc5\xdb\x7f\x3f"
 ,"\x0e\x00\x1a\xfa\x30\x40\x9c\x46\x1b\x35\xc7\xd9\xc2\xb3\x47\x2a\x81\x31\xa4\x11\x8a\x2d\x02\xaa\x32\xbd\xa0\xc1\x7f\x3f"
 ,"\x04\x00\xb0\xd6\x9b\xb5\xee\x7f\x40\x3b\x8d\xd5\x92\xbf\xe3\x78\x22\xbf\x1d\x36\xd4\x3d\xa2\x7c\x41\x3f\x9c\xb9\xfc\xbd"
 ,"\x10\x00\x20\xed\xa8\x40\x97\xa3\x05\xb4\x47\xe5\x54\xb5\x8d\x52\xc3\x3c\x8d\x87\x04\xbd\xb4\x95\xdc\x3c\x44\xb3\x7f\x3f"
 ,"\x11\x00\xbe\xde\x1b\x41\xa0\xc4\x6a\xb4\x85\x4a\x8f\x35\x3f\xcc\x13\xb1\x34\x02\xad\xb0\xee\x0d\x2d\x3e\x5a\x51\x7c\x3f"
 ,"\x12\x00\x0d\x0a\x21\x41\x48\x1a\x92\xb4\x09\xa1\x47\xb4\xff\xae\x7f\x3f\xf5\x37\x20\x3d\x7a\x55\xf8\xbc\xc8\xe2\x97\xbb"
 ,"\x13\x00\x4b\xcb\x0f\x3f\x07\xd4\xee\x3e\x35\x1c\x89\x3f\x49\xbb\x30\x3f\x2d\x8b\xae\x3e\x7f\xc0\x1c\x3e\xfe\x95\x1e\x3f"
 ,"\x14\x00\xf9\x37\xab\x3f\x2a\x05\x32\x36\x91\x42\xa1\xb4\xf0\xa5\xef\x31\x65\x2e\x4f\xb3\x2e\x02\x29\xbe\x37\x7d\x7c\x3f"
 ,"\x13\x00\xe8\xf9\x39\x40\x72\x7d\x2d\xbc\xa5\xcb\x3e\x3f\xae\x4a\xb5\xbb\x38\x0f\x56\x3d\x76\xa5\x0c\xbd\xbe\x7e\x7f\x3f"
 ,"\x16\x00\x23\xdd\x2d\x40\xf9\x00\x58\xb6\x18\xc7\x8f\x34\x2d\x2d\x35\xb2\xf3\x92\x7e\x31\xad\xf2\x0e\xbd\x14\xd8\x7f\x3f"
 ,"\x13\x00\x89\x60\x3b\x40\x6b\x13\xc5\xba\x4c\x6b\x06\xbf\x75\x95\x7c\xbb\xfe\xbb\x87\xb5\x3d\x73\xf8\xbc\x5e\xe1\x7f\x3f"
 ,"\x18\x00\x0e\xfa\x30\x40\x09\x07\x52\x36\xdc\x87\x19\xb4\x28\xa7\x8a\xb1\xfe\x20\x0b\x32\x05\xaa\x32\xbd\xa0\xc1\x7f\x3f"
 ,"\x00\x00\x11\x90\x94\xb6\x2b\x6c\xc7\x36\x4a\xd3\x47\x40\xec\x69\x00\xbd\xc9\xdf\x7f\x3f\x33\x26\x34\xb5\xf5\x92\x34\xb5"
 ,"\x1a\x00\x2c\x12\x9a\x41\x30\x56\x93\x31\xaf\xdd\x58\xb3\xfb\x68\x8b\xa8\xf7\xd2\xe6\xa7\xc4\xe7\x83\x3a\xf8\xff\x7f\x3f"
 ,"\x1b\x00\x18\x42\x5c\x41\x70\xdd\x58\x32\xab\x4d\x3c\x34\x06\xa4\x3b\xb3\xbf\x62\xc2\xb0\xa3\x88\x04\xbd\xaf\xdd\x7f\x3f"
 ,"\x1c\x00\xcf\xe1\x86\x40\xe7\x93\x4c\x40\x72\xac\xc9\x33\x09\x06\x35\x29\xbd\xa6\xdd\xa8\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
 ,"\x00\x00\x2e\x04\x8e\x36\xa0\xcf\x34\xb6\x4a\xd3\x47\xc0\xec\x69\x00\xbd\xc9\xdf\x7f\x3f\x33\x26\x34\xb5\xf5\x92\x34\xb5"
 ,"\x1e\x00\x2c\x12\x9a\x41\x9c\x55\x04\x34\xca\xc5\x58\xb3\xfb\x68\x8b\xa8\xf7\xd2\xe6\xa7\xc4\xe7\x83\x3a\xf8\xff\x7f\x3f"
 ,"\x1f\x00\x17\x42\x5c\x41\x80\x9e\x28\x32\xa4\x53\x3c\x34\x06\xa4\x3b\xb3\xbf\x62\xc2\xb0\xa3\x88\x04\xbd\xaf\xdd\x7f\x3f"
 ,"\x20\x00\xd1\xe1\x86\x40\xe7\x93\x4c\x40\xe6\xb8\xc9\x33\x09\x06\x35\x29\xbd\xa6\xdd\xa8\xf3\x04\x35\xbf\xf3\x04\x35\x3f"
};
#endif

#endif
