#include "gsmhacking.h"

struct gripper **alloc_grippers(unsigned short n) {
  return (struct gripper **) alloc_lots(n, "gripper", sizeof(struct gripper *), (void *(*)()) alloc_gripper);
}

struct gripper *alloc_gripper() {
  return (struct gripper *) alloc_one("gripper", sizeof(struct gripper));
}

void cleanup_grippers(unsigned short num_grippers, struct gripper **grippers) {
  unsigned short i;

  if (grippers) {
    for (i = 0; i < num_grippers; i++) {
      DEBUG("freeing %u %s\n", i, grippers[i]->name);
      if (grippers[i]->name) free(grippers[i]->name);
    }
    free_structs("gripper", num_grippers, (void *) grippers);
  }
}

void cleanup_grips(unsigned short num_grips, struct gripper **grips) {
  unsigned short i;

  if (grips) {
    for (i = 0; i < num_grips; i++) {
      DEBUG("freeing %u %s\n", i, grips[i]->name);
      if (grips[i]->name) free(grips[i]->name);
    }
    free_structs("grip", num_grips, (void *) grips);
  }
}

/* grips and grippers are the same for all practical purposes */
int find_grips(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct gripper *grip;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think grips start at 0x%04x\n", pos);

  /* how many grips are there? */
  if (read_short(gsm->fd, &gsm->num_grips)) {
    ERROR("can't read number of grips");
    return 2;
  }
  DEBUG("grips: %u\n", gsm->num_grips);
  if (! gsm->num_grips) return 0;

  /* allocate a big enough struct */
  gsm->grips = alloc_grippers(gsm->num_grips);
  if (! gsm->grips) return 4;

  /* fill in the grips */
  for (i = 0; i < gsm->num_grips; i++) {
    grip = gsm->grips[i];
    grip->name = 0;
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    grip->offset = pos;
    if (parse_gripper(gsm, i, grip, gsm->num_grips)) {
      ERROR("expected grip id %u but got null\n", i);
      return 6;
    }
  }

  return 0;
}

int find_grippers(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct gripper *gripper;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think grippers start at 0x%04x\n", pos);

  /* how many grippers are there? */
  if (read_short(gsm->fd, &gsm->num_grippers)) {
    ERROR("can't read number of grippers");
    return 2;
  }
  if (! gsm->num_grippers) {
    if (! gsm->grips) {
      SAY("%s has no grips or grippers\n", gsm->filename);
    }
    return 0;
  }
  DEBUG("grippers: %u\n", gsm->num_grippers);

  /* allocate a big enough struct */
  gsm->grippers = alloc_grippers(gsm->num_grippers);
  if (! gsm->grippers) return 4;

  /* fill in the grippers */
  for (i = 0; i < gsm->num_grippers; i++) {
    gripper = gsm->grippers[i];
    gripper->name = 0;
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    gripper->offset = pos;
    if (parse_gripper(gsm, i, gripper, gsm->num_grippers)) {
      ERROR("expected gripper id %u but got null\n", i);
      return 6;
    }
  }

  return 0;
}

int parse_gripper(struct gsm *gsm, unsigned short num, struct gripper *gripper, unsigned short total) {
  /* find name length */
  if (read_short(gsm->fd, &gripper->name_length)) {
    ERROR("can't read length of gripper name");
    return 2;
  }
  TRACE("name is of length %u\n", gripper->name_length);

  /* read name */
  gripper->name = (char *) malloc((sizeof(char) * gripper->name_length + 1));
  if (! gripper->name) {
    ERROR("no memory for gripper name");
    return 3;
  }
  if (read_arbitrary(gsm->fd, gripper->name, gripper->name_length)) {
    ERROR("can't read name of gripper");
    return 4;
  }
  (gripper->name)[gripper->name_length] = '\0';
  DEBUG("gripper: %s\n", gripper->name);

  /* read transform matrix */
  if (read_matrix(gsm->fd, &gripper->rotation)) {
    ERROR("can't read gripper transformation matrix");
    return 5;
  }
  debug_matrix(&gripper->rotation);
  /* find position vector */
  if (read_vector(gsm->fd, &gripper->position)) {
    ERROR("can't read gripper co-ordinates");
    return 6;
  }

  /* find bone */
  if (read_short(gsm->fd, &gripper->bone_id)) {
    ERROR("can't read gripper bone ID");
    return 7;
  }
  TRACE("gripper %u at 0x%04x: %s grips %02u %s (% 0.4f, % 0.4f, % 0.4f)\n", num, gripper->offset, gripper->name, gripper->bone_id, get_bone_by_id(gsm, gripper->bone_id), gripper->position.x, gripper->position.y, gripper->position.z);

  return 0;
}

/* make standard SWT grippers (for conversion) */
struct gripper **make_swt_grippers() {
  unsigned short i;
  struct gripper **grippers;
 
  grippers = alloc_grippers(SWT_NUM_GRIPPERS);
  if (! grippers) return 0;

  for (i = 0; i < SWT_NUM_GRIPPERS; i++) {
    if (make_swt_gripper(i, grippers)) {
      ERROR("can't make gripper #%u\n", i);
      cleanup_grippers(SWT_NUM_GRIPPERS, grippers);
      return 0;
    }
  }

  return grippers;
}

int make_swt_gripper(unsigned short i, struct gripper **grippers) {
  grippers[i]->offset = 0x0fc0;
  grippers[i]->name_length = 0;
  grippers[i]->name = alloc_name(get_gripper_name(GSM_TYPE_SWT, i), &grippers[i]->name_length);
  if (! grippers[i]->name) {
    ERROR("can't set gripper #%u name\n", i);
    return 1;
  }
  TRACE2("gripper %02u new name: %s\n", i, grippers[i]->name);

  TRACE2("setting gripper %02u co-ordinates\n", i);
  memmove(&grippers[i]->rotation[R0], swt_gripper_coords[i], sizeof(float));
  memmove(&grippers[i]->rotation[R1], swt_gripper_coords[i] + 4, sizeof(float));
  memmove(&grippers[i]->rotation[R2], swt_gripper_coords[i] + 8, sizeof(float));
  memmove(&grippers[i]->rotation[R3], swt_gripper_coords[i] + 12, sizeof(float));
  memmove(&grippers[i]->rotation[R4], swt_gripper_coords[i] + 16, sizeof(float));
  memmove(&grippers[i]->rotation[R5], swt_gripper_coords[i] + 20, sizeof(float));
  memmove(&grippers[i]->rotation[R6], swt_gripper_coords[i] + 24, sizeof(float));
  memmove(&grippers[i]->rotation[R7], swt_gripper_coords[i] + 28, sizeof(float));
  memmove(&grippers[i]->rotation[R8], swt_gripper_coords[i] + 32, sizeof(float));
  memmove(&grippers[i]->position.x, swt_gripper_coords[i] + 36, sizeof(float));
  memmove(&grippers[i]->position.y, swt_gripper_coords[i] + 40, sizeof(float));
  memmove(&grippers[i]->position.z, swt_gripper_coords[i] + 44, sizeof(float));

  TRACE2("setting gripper %02u bone number\n", i);
  grippers[i]->bone_id = swt_gripper_bone_id[i];

  return 0;
}


/* make standard TAC grippers (for conversion) */
struct gripper **make_tac_grippers() {
  unsigned short i;
  struct gripper **grippers;
 
  grippers = alloc_grippers(TAC_NUM_GRIPPERS);
  if (! grippers) return 0;

  for (i = 0; i < TAC_NUM_GRIPPERS; i++) {
    if (make_tac_gripper(i, grippers)) {
      ERROR("can't make gripper #%u\n", i);
      cleanup_grippers(TAC_NUM_GRIPPERS, grippers);
      return 0;
    }
  }

  return grippers;
}

int make_tac_gripper(unsigned short i, struct gripper **grippers) {
  grippers[i]->offset = 0x0fc0;
  grippers[i]->name_length = 0;
  grippers[i]->name = alloc_name(get_gripper_name(GSM_TYPE_TAC, i), &grippers[i]->name_length);
  if (! grippers[i]->name) {
    ERROR("can't set gripper #%u name\n", i);
    return 1;
  }
  TRACE2("gripper %02u new name: %s\n", i, grippers[i]->name);

  TRACE2("setting gripper %02u co-ordinates\n", i);
  memmove(&grippers[i]->rotation[R0], tac_gripper_coords[i], sizeof(float));
  memmove(&grippers[i]->rotation[R1], tac_gripper_coords[i] + 4, sizeof(float));
  memmove(&grippers[i]->rotation[R2], tac_gripper_coords[i] + 8, sizeof(float));
  memmove(&grippers[i]->rotation[R3], tac_gripper_coords[i] + 12, sizeof(float));
  memmove(&grippers[i]->rotation[R4], tac_gripper_coords[i] + 16, sizeof(float));
  memmove(&grippers[i]->rotation[R5], tac_gripper_coords[i] + 20, sizeof(float));
  memmove(&grippers[i]->rotation[R6], tac_gripper_coords[i] + 24, sizeof(float));
  memmove(&grippers[i]->rotation[R7], tac_gripper_coords[i] + 28, sizeof(float));
  memmove(&grippers[i]->rotation[R8], tac_gripper_coords[i] + 32, sizeof(float));
  memmove(&grippers[i]->position.x, tac_gripper_coords[i] + 36, sizeof(float));
  memmove(&grippers[i]->position.y, tac_gripper_coords[i] + 40, sizeof(float));
  memmove(&grippers[i]->position.z, tac_gripper_coords[i] + 44, sizeof(float));

  TRACE2("setting gripper %02u bone number\n", i);
  grippers[i]->bone_id = tac_gripper_bone_id[i];

  return 0;
}


/* make standard SWT grippers (for conversion) */
struct gripper **make_bf1_grippers() {
  unsigned short i;
  struct gripper **grippers;
 
  grippers = alloc_grippers(BF1_NUM_GRIPPERS);
  if (! grippers) return 0;

  for (i = 0; i < BF1_NUM_GRIPPERS; i++) {
    if (make_bf1_gripper(i, grippers)) {
      ERROR("can't make gripper #%u\n", i);
      cleanup_grippers(BF1_NUM_GRIPPERS, grippers);
      return 0;
    }
  }

  return grippers;
}

int make_bf1_gripper(unsigned short i, struct gripper **grippers) {
  grippers[i]->offset = 0x0fc0;
  grippers[i]->name_length = 0;
  grippers[i]->name = alloc_name(get_gripper_name(GSM_TYPE_BF1, i), &grippers[i]->name_length);
  if (! grippers[i]->name) {
    ERROR("can't set gripper #%u name\n", i);
    return 1;
  }
  TRACE2("gripper %02u new name: %s\n", i, grippers[i]->name);

  TRACE2("setting gripper %02u co-ordinates\n", i);
  memmove(&grippers[i]->rotation, &bf1_gripper_rotation[i], sizeof(matrix_t));
  memmove(&grippers[i]->position, &bf1_gripper_position[i], sizeof(vector_t));

  TRACE2("setting gripper %02u bone number\n", i);
  grippers[i]->bone_id = bf1_gripper_bone_id[i];

  return 0;
}

int write_grips(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing grips: %u\n", gsm->num_grips);
  if (write_short(gsm->out_fd, &gsm->num_grips)) {
    ERROR("can't write number of grips");
    return 3;
  }

  for (i = 0; i < gsm->num_grips; i++) {
    if (write_gripper(gsm->out_fd, gsm->grips[i])) {
      ERROR("can't write grip %u\n", i);
      return 4;
    }
  }

  return 0;
}

int write_grippers(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing grippers: %u\n", gsm->num_grippers);
  if (write_short(gsm->out_fd, &gsm->num_grippers)) {
    ERROR("can't write number of grippers");
    return 3;
  }

  for (i = 0; i < gsm->num_grippers; i++) {
    if (write_gripper(gsm->out_fd, gsm->grippers[i])) {
      ERROR("can't write gripper %u\n", i);
      return 4;
    }
  }

  return 0;
}

int write_gripper(FILE *fd, struct gripper *gripper) {
  if (! gripper) {
    ERROR("no gripper!\n");
    return 1;
  }

  TRACE("writing gripper name \"%s\" of length %u\n", gripper->name, gripper->name_length);
  if (write_short(fd, &gripper->name_length)) {
    ERROR("can't write gripper name length");
    return 2;
  }
  if (write_arbitrary(fd, gripper->name, gripper->name_length)) {
    ERROR("can't write gripper name");
    return 3;
  }
  if (write_matrix(fd, &gripper->rotation)) {
    ERROR("can't write gripper transformation matrix");
    return 4;
  }
  if (write_vector(fd, &gripper->position)) {
    ERROR("can't write gripper co-ordinates");
    return 5;
  }
  if (write_short(fd, &gripper->bone_id)) {
    ERROR("can't write gripper bone id");
    return 6;
  }

  return 0;
}

char *get_gripper_name(unsigned short type, unsigned short number) {
  if (type == GSM_TYPE_SWT) {
    if (number <= SWT_NUM_GRIPPERS) return swt_gripper_name[number];
    return "BOGUS";
  }
  if (type == GSM_TYPE_TAC) {
    if (number <= TAC_NUM_GRIPPERS) return tac_gripper_name[number];
    return "BOGUS";
  }
  if (type == GSM_TYPE_BF1) {
    if (number <= BF1_NUM_GRIPPERS) return bf1_gripper_name[number];
    return "BOGUS";
  }
  /* FIXME: guns and throwables */
  return "FIXME";
}

int get_gripper_by_bone(struct gsm *gsm, unsigned short bone_id) {
  unsigned short i;

  for (i = 0; i < gsm->num_grippers; i++) {
    if (gsm->grippers[i]->bone_id == bone_id) return i;
  }

  return -1;
}

int get_grip_by_bone(struct gsm *gsm, unsigned short bone_id) {
  unsigned short i;

  for (i = 0; i < gsm->num_grips; i++) {
    if (gsm->grips[i]->bone_id == bone_id) return i;
  }

  return -1;
}

int describe_grippers(struct gsm *gsm) {
  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  return describe_gripper_objects("Gripper", gsm, gsm->grippers, gsm->num_grippers);
}

int describe_grips(struct gsm *gsm) {
  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  return describe_gripper_objects("Grip", gsm, gsm->grips, gsm->num_grips);
}

int describe_gripper_objects(char *name, struct gsm *gsm, struct gripper **gripper_objects, unsigned short num_gripper_objects) {
  unsigned short i;
  struct gripper *gripper;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (gsm_verbose) SAY("%ss: %u\n", name, num_gripper_objects);
  if (! num_gripper_objects) return 0;
  if (! gripper_objects) {
    ERROR("null gripper pointer\n");
    return 2;
  }

  for (i = 0; i < num_gripper_objects; i++) {
    gripper = gripper_objects[i];
    if (! gsm_verbose) {
      SAY("%s %02u at 0x%04x: %s grips %s\n", name, i, gripper->offset, gripper->name, get_bone_by_id(gsm, gripper->bone_id));
      continue;
    }
    SAY("  %s %02u at 0x%04x: %s\n", name, i, gripper->offset, gripper->name);
    SAY("  Position:  (% 0.4f, % 0.4f, % 0.4f)\n", gripper->position.x, gripper->position.y, gripper->position.z);
    SAY("  Transform: (% 0.4f, % 0.4f, % 0.4f)\n             (% 0.4f, % 0.4f, % 0.4f)\n             (% 0.4f, % 0.4f, % 0.4f)\n", gripper->rotation[R0], gripper->rotation[R3], gripper->rotation[R6], gripper->rotation[R1], gripper->rotation[R4], gripper->rotation[R7], gripper->rotation[R2], gripper->rotation[R5], gripper->rotation[R8]);
    SAY("  Grips: %s\n", get_bone_by_id(gsm, gripper->bone_id));
    SAY("\n");
  }
  if (! num_gripper_objects) SAY("\n");

  return 0;
}
