#ifndef GRIPPER_H
#define GRIPPER_H
#include <stdlib.h>
#include "data.h"

#define GRIPPER_COORDS_LENGTH 48
#define SWT_NUM_GRIPPERS 3
#define TAC_NUM_GRIPPERS 3
#define BF1_NUM_GRIPPERS 3
#define GUN_NUM_GRIPS 2
#define PIS_NUM_GRIPS 1

GSM_API struct gripper **alloc_grippers(unsigned short);
GSM_API struct gripper *alloc_gripper();
GSM_API void cleanup_grippers(unsigned short, struct gripper **);
GSM_API void cleanup_grips(unsigned short, struct gripper **);
GSM_API int find_grips(struct gsm *);
GSM_API int find_grippers(struct gsm *);
GSM_API int parse_gripper(struct gsm *, unsigned short, struct gripper *, unsigned short);
GSM_API struct gripper **make_swt_grippers();
GSM_API int make_swt_gripper(unsigned short, struct gripper **);
GSM_API struct gripper **make_tac_grippers();
GSM_API int make_tac_gripper(unsigned short, struct gripper **);
GSM_API struct gripper **make_bf1_grippers();
GSM_API int make_bf1_gripper(unsigned short, struct gripper **);
GSM_API int write_grips(struct gsm *);
GSM_API int write_grippers(struct gsm *);
GSM_API int write_gripper(FILE *, struct gripper *);
GSM_API char *get_gripper_name(unsigned short, unsigned short);
GSM_API int get_gripper_by_bone(struct gsm *, unsigned short);
GSM_API int get_grip_by_bone(struct gsm *, unsigned short);
GSM_API int describe_grippers(struct gsm *);
GSM_API int describe_grips(struct gsm *);
GSM_API int describe_gripper_objects(char *, struct gsm *, struct gripper **, unsigned short);

struct gripper {
  unsigned int offset;
  unsigned short name_length;
  char *name;
  vector_t position;
  matrix_t rotation;
  unsigned short bone_id;
};

/* damn gonna have to do this properly */
#ifndef __cplusplus
/* SWT */
static unsigned char swt_gripper_coords[SWT_NUM_GRIPPERS][GRIPPER_COORDS_LENGTH] = {
  "\x90\x6d\x7c\x3f\x24\x1b\xcd\x3d\xd5\x26\x08\x3e\xa9\x95\x00\xbe\x98\x8c\x9d\xbd\x1d\x36\x7d\x3f\xc1\xd1\xdf\x3d\x52\xf3\x7d\xbf\xa3\x58\x81\xbd\xe3\xec\x1c\x40\x49\x02\x79\x3f\x36\x39\xd6\x3f"
 ,"\xdf\x97\x5b\x3f\xed\x09\xeb\x3e\x00\xd6\x6c\x3e\xf9\xb4\xe5\xbe\x7e\x61\x63\x3f\x3c\x94\xca\xbd\xbd\x6d\x80\xbe\x8d\xfe\x99\xbc\xa6\xc4\x77\x3f\x6a\xc5\x85\x40\xdd\xa6\xe8\x3f\x48\x7c\x40\xc0"
 ,"\xc8\xb0\x2c\xbf\xa1\x60\x9f\xbe\xc2\x5b\x2b\x3f\x67\x10\x3b\x3f\xf2\x63\x1c\xbe\x12\x56\x2a\x3f\x70\xd1\xd6\xbd\x6e\x1e\x70\x3f\x3f\x35\xa9\x3e\xe7\x94\xb1\x40\xb1\x45\x20\x41\xa2\x10\x11\x40"
};
static unsigned short swt_gripper_bone_id[SWT_NUM_GRIPPERS] = { 28, 9, 1 };
static unsigned char *swt_gripper_name[SWT_NUM_GRIPPERS] = {
  "Grip1", "Grip2", "Grip3"
};

/* TAC, BG1, BG2, BG3, MH1, MH2 */
static unsigned char tac_gripper_coords[TAC_NUM_GRIPPERS][GRIPPER_COORDS_LENGTH] = {
  "\x42\xeb\x7c\x3f\x94\xc6\x0d\x3e\xe1\x35\x8d\x3d\x15\xbe\x92\xbd\xc3\x51\x8c\x3c\xc6\x4d\x7f\x3f\x7d\x31\x2f\x3e\x4b\x19\x7c\xbf\xa2\x64\x00\x3d\xdf\xac\x2c\x40\x9b\xf6\x17\x3f\x7a\x02\x98\x3f"
 ,"\xce\x59\x76\x3f\xcd\x49\x5d\x3e\x73\x1b\x29\x3e\x2e\xec\x55\xbe\x5c\xe1\x79\x3f\x04\x20\x75\xbd\xc6\x4e\x32\xbe\x38\x23\xbe\x3c\x0f\x05\x7c\x3f\x08\xaa\x8c\x40\x47\xd1\x60\x3f\xd4\x30\x3e\xc0"
 ,"\xc9\xb0\x2c\xbf\xa1\x60\x9f\xbe\xc2\x5b\x2b\x3f\x69\x10\x3b\x3f\xf2\x63\x1c\xbe\x10\x56\x2a\x3f\x64\xd1\xd6\xbd\x6f\x1e\x70\x3f\x3c\x35\xa9\x3e\xe7\x94\xb1\x40\xb1\x45\x20\x41\xa2\x10\x11\x40"
};
static unsigned short tac_gripper_bone_id[TAC_NUM_GRIPPERS] = { 22, 12, 1 };
static unsigned char *tac_gripper_name[TAC_NUM_GRIPPERS] = {
  "Grip1", "Grip2", "Grip3"
};

/* BF1, FH1, FH2 */
static vector_t bf1_gripper_position[BF1_NUM_GRIPPERS] = {
  {  2.2503f,  1.0146f, -0.9422f }
 ,{  1.6644f,  1.3787f, -2.2697f }
 ,{  5.4517f,  7.3423f,  2.2609f }
};
static matrix_t bf1_gripper_rotation[BF1_NUM_GRIPPERS] = {
  {  0.9985f, -0.0368f, -0.0401f,  0.0000f, -0.0390f,  0.0012f, -0.9992f,  0.0000f,   0.344f,  0.9991f, -0.0259f,  0.0000f,  0.0000f,  0.0000f,  0.0000f }
 ,{ -0.9004f, -0.3496f,  0.2589f,  0.0000f,  0.3351f, -0.9369f, -0.0997f,  0.0000f,  0.2774f, -0.0029f,  0.9607f,  0.0000f,  0.0000f,  0.0000f,  0.0000f }
 ,{ -0.6838f, -0.2483f, -0.6861f,  0.0000f,  0.7296f, -0.2392f, -0.6407f,  0.0000f, -0.0050f, -0.9387f,  0.3447f,  0.0000f,  0.0000f,  0.0000f,  0.0000f }
};
static unsigned short bf1_gripper_bone_id[BF1_NUM_GRIPPERS] = { 9, 19, 1 };
static unsigned char *bf1_gripper_name[BF1_NUM_GRIPPERS] = {
  "Grip1", "Grip2", "Grip3"
};

/* weapons */
static unsigned char *gun_grip_name[GUN_NUM_GRIPS] = {
  "Grip1", "Grip3"
};
#endif

#endif
