#ifndef GSM_H
#define GSM_H
#include <stdlib.h>
#include <stdio.h>
#include "option.h"

#define GSM_MAGIC "\x72\xf6\x63\x1e\x5e\x47\xd2\x11\x87\x99\x00\x20\xaf\xe6\x36\xee"
#define GSM_MAGIC_LEN 16

#define MUZZLE "MUZZLE"
#define FLASHLIGHT "FLASHLIGHT"

GSM_API struct gsm **alloc_gsms(unsigned short);
GSM_API struct gsm *alloc_gsm();
GSM_API int read_gsm_file(struct gsm *);
GSM_API int write_gsm_file(struct gsm *);
GSM_API int open_gsm_file(struct gsm *);
GSM_API int write_gsm_data(struct gsm *);
GSM_API int link_gsm_file(struct gsm *);
GSM_API int close_gsm_file(struct gsm *);
GSM_API int write_magic(struct gsm *);
GSM_API int check_magic(struct gsm *, char *);
GSM_API void cleanup_gsm(struct gsm *);
GSM_API int parse_gsm(struct gsm *);
GSM_API void classify_gsm(struct gsm *);
GSM_API void show_gsm(struct gsm *);
GSM_API void show_polycount(struct gsm *);
GSM_API void summarise_gsm(struct gsm *);
GSM_API int describe_gsm(struct gsm *);

#define GSM_TYPE_UNKNOWN 0
#define GSM_TYPE_GUN 1
#define GSM_TYPE_SWT 2
#define GSM_TYPE_TAC 4
#define GSM_TYPE_BF1 8
#define GSM_TYPE_MHx 16
#define GSM_TYPE_FHx 32
#define GSM_TYPE_KID 64
#define GSM_TYPE_RAT 128
#define GSM_TYPE_FISH 256
#define GSM_TYPE_OBJ 512
#define GSM_TYPE_MAG 1024
#define GSM_TYPE_PIS 2048
#define GSM_TYPE_TOOL 4096
#define GSM_TYPE_DEFAULT 8192 /* default skeleton */
#define GSM_TYPE_ANIMATION_LIMITED_BONES 16384
#define GSM_TYPE_ANIMATION 32768 /* bitwase OR with this to denote animation */

struct gsm {
  char *filename;
  char *basename;
  char *outfile;
  char *out_basename;
  char *tmpfile;
  struct option *option;
  unsigned short basename_length;
  unsigned short out_basename_length;
  FILE *fd;
  FILE *out_fd;
  unsigned short version;
  unsigned short num_skeletons;
  unsigned short num_assignments;
  unsigned short num_positions;
  unsigned short num_blended_positions;
  unsigned short num_texture_coords;
  unsigned short num_vertices;
  unsigned short num_triangles;
  unsigned short num_triangle_groups;
  unsigned short num_shared_triangle_lists;
  unsigned short num_textures;
  unsigned short num_materials;
  unsigned short num_grips;
  unsigned short num_grippers;
  unsigned short num_frames; /* for animation */
  unsigned char need_assignments_hack;
  unsigned int shared_triangle_list_length;
  short first_blended_position;
  int number_blended_positions; /* int, num_blended_positions is short! */
  unsigned int blended_position_stream_length;
  struct skeleton **skeletons;
  struct assignment **assignments;
  struct position **positions;
  struct blended_position **blended_positions;
  struct texture_coord **texture_coords;
  struct vertex **vertices;
  struct triangle **triangles;
  struct triangle_group **triangle_groups;
  struct shared_triangle_list **shared_triangle_lists;
  struct texture **textures;
  struct material **materials;
  struct gripper **grips;
  struct gripper **grippers;
  struct frame **frames;
  unsigned short type;
  unsigned char is_animation;
  /* optimisations */
  unsigned short unmatched_materials;
  unsigned short unmatched_textures;
  unsigned char need_make_rifle;
  unsigned char need_optimise;
  /* animation stuff */
  unsigned short animation_unknown_before_framerate;
  unsigned short framerate;
  unsigned short animation_unknown_after_framerate;
};

#endif
