#ifdef USE_IMAGEMAGICK
#include "gsmhacking.h"

struct generate_uv_map_set *alloc_generate_uv_map_set() {
  struct generate_uv_map_set *ret = (struct generate_uv_map_set *) alloc_one("generate_uv_map_set", sizeof(struct generate_uv_map_set));
  if (! ret) return 0;
  return ret;
}

void cleanup_generate_uv_map_sets(struct generate_uv_map_set *generate_uv_map_sets) {
  struct generate_uv_map_set *tmp;
  SAY("freeing generate_uv_map_sets at %p\n", generate_uv_map_sets);
  gll_free((struct gll *) generate_uv_map_sets);
}

int add_to_generate_uv_map_set(struct generate_uv_map_set **generate_uv_map_sets, char *group, char *colour) {
  struct generate_uv_map_set *last, *tmp;
  int triangle_group_id;
  char *bogus;
  int i, ok;
  unsigned short len;
 
  tmp = alloc_generate_uv_map_set();
  if (! tmp) {
    ERROR("out of memory for generate_uv_map_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* get the group */
  triangle_group_id = strtol(group, &bogus, 0);
  if (*bogus || triangle_group_id < 0) {
    ERROR("bogus group ID: try a non-negative number\n");
    return 2;
  }
  tmp->triangle_group_id = (unsigned short) triangle_group_id;
  TRACE("tmp->triangle_group_id: %u\n", tmp->triangle_group_id);

  /* get the colour */
  tmp->wand = NewPixelWand();
  if (str_equiv(colour, "transparent")) {
    tmp->transparent = 1;
    PixelSetColor(tmp->wand, "white");
    TRACE("tmp->transparent: 1 (white on black)\n");
  }
  else {
    PixelSetColor(tmp->wand, colour);
    bogus = PixelGetColorAsString(tmp->wand);
    if (! bogus) {
      ERROR("bogus colour string '%s': try #xxxxxx (like HTML)\n", colour);
      return 3;
    }
    TRACE("tmp->colour: %s\n", bogus);
  }

  last  = (struct generate_uv_map_set *) gll_last((struct gll *) *generate_uv_map_sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *generate_uv_map_sets = tmp;
  /*gll_debug((struct gll *) *generate_uv_map_sets);*/

  return 0;
}

int generate_uv_maps(struct gsm *gsm) {
  struct generate_uv_map_set *generate_uv_map_set, *generate_uv_map_sets;
  int errors, ret;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  if (gsm->is_animation) {
    ERROR("animations don't have UVs\n");
    return -1;
  }

  generate_uv_map_sets = gsm->option->generate_uv_map_sets;
  if (! generate_uv_map_sets) {
    ERROR("no generate_uv_map_sets\n");
    return 2;
  }

  if (IsMagickInstantiated() == MagickFalse) {
    MagickCoreGenesis(0, MagickTrue);
    MagickWandGenesis();
    DEBUG("magick initialised\n");
  }
  errors = 0;
  GLL_LOOP(generate_uv_map_set, generate_uv_map_sets) {
    ret = generate_uv_map(gsm, generate_uv_map_set);
    if (ret) {
      ERROR("generate_uv_map() returned %d\n", ret);
      errors++;
    }
  }
  MagickCoreTerminus();
  DEBUG("magick finalised\n");

  return errors;
}

int generate_uv_map(struct gsm *gsm, struct generate_uv_map_set *set) {
  struct gsm *src;
  /* here begins the ImageMagick stuff */
  ExceptionInfo oops;
  Image *image;
  MagickWand *image_wand;
  DrawingWand *wand;
  ImageInfo image_info;
  static char pixels[256 * 256 * 3];
  /* here ends the ImageMagick stuff */
  struct triangle_group *triangle_group;
  struct material *material;
  struct map *map;
  struct texture *texture;
  struct triangle *triangle;
  struct vertex *vertex;
  struct texture_coord **uv;
  struct stat st;
  char *filename;
  unsigned short i;
  int x0, y0, x1, y1;
  double tmp;
  char *ext;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! set) {
    ERROR("no generate_uv_map_set\n");
    return 2;
  }

  /* check group ID */
  if (set->triangle_group_id >= gsm->num_triangle_groups) {
    ERROR("triangle group ID %u out of range (max %u)\n", set->triangle_group_id, gsm->num_triangle_groups);
    return 3;
  }

  /* get texture */
  triangle_group = gsm->triangle_groups[set->triangle_group_id];
  if (triangle_group->material_id >= gsm->num_materials) {
    ERROR("huh? material ID out of range!\n");
    return 4;
  }
  material = gsm->materials[triangle_group->material_id];
  /* don't bother looking for a texture if we want transparency */
  if (set->transparent) {
    filename = 0;
    ext = "bmp";
  }
  else {
    if (! material->num_maps) {
      INFO("group %u has material %u which has no maps\n");
      /* force transparency */
      filename = 0;
      ext = "bmp";
    }
    else {
      /* use the first map */
      map = material->maps[0];
      if (map->texture_id >= gsm->num_textures) {
        ERROR("huh? texture ID out of range!\n");
        return 5;
      }
      texture = gsm->textures[map->texture_id];
  
      /* check we can load that sucker */
      if (! stat(texture->basename, &st)) filename = texture->basename;
      else if (! stat(texture->filename, &st)) filename = texture->filename;
      else {
        WARN("texture file %s not found; will create a blank image\n", texture->basename);
        ext = "bmp";
        filename = 0;
      }
    }
  }

  /* start magick */
  GetImageInfo(&image_info);
  GetExceptionInfo(&oops);

  if (filename) {
    DEBUG("reading image from %s\n", filename);
    if (snprintf(image_info.filename, MaxTextExtent, "%s", filename) == -1) {
      ERROR("snprintf() barfed");
      return 5;
    }

    image = ReadImage(&image_info, &oops);
    if (! image) {
      ERROR("failed to load texture %s: %s (%s)\n", filename, oops.reason, oops.description);
      return 6;
    }

    /* get file extension for later */
    for (i = 0; *(filename + i); i++);
    for ( ; i; i--) {
      if (*(filename + i) == '.') {
        i++;
        break;
      }
    }
    ext = filename + i;
  }
  else {
    /* make new image */
    memset(pixels, 0, sizeof(pixels));
    DEBUG("creating blank image\n");
    image = ConstituteImage(256, 256, "RGB", CharPixel, pixels, &oops);
    if (! image) {
      ERROR("can't generate new image: %s (%s)\n", oops.reason, oops.description);
      return 7;
    }
    /* white on black */
    PixelSetColor(set->wand, "white");
  }
  DEBUG("image acquired\n");

  /* create the wands */
  wand = NewDrawingWand();
  image_wand = NewMagickWandFromImage(image);
  if (! image_wand) {
    ERROR("failed to allocate wand from image!\n");
    DestroyImage(image);
    return 8;
  }

  /* set the image dimensions and colours */
  DrawSetViewbox(wand, 0, 0, image->columns, image->rows);
  DrawScale(wand, 1.0, 1.0);
  DrawRotate(wand, 0.0);
  DrawSetStrokeAntialias(wand, MagickTrue);
  DrawSetStrokeWidth(wand, 1.0);
  DrawSetStrokeColor(wand, set->wand);

  /* doit */
  uv = (struct texture_coord **) malloc(3 * sizeof(struct texture_coord *));
  if (! uv) {
    ERROR("out of memory for temporary UVs\n");
    return 10;
  }
  for (i = 0; i < triangle_group->num_triangles; i++) {
    triangle = gsm->triangles[i + triangle_group->first_triangle];

    /* get UVs for each point of the tri */
    vertex = gsm->vertices[triangle->vertex1];
    uv[0] = gsm->texture_coords[vertex->texture_coord_id];
    vertex = gsm->vertices[triangle->vertex2];
    uv[1] = gsm->texture_coords[vertex->texture_coord_id];
    vertex = gsm->vertices[triangle->vertex3];
    uv[2] = gsm->texture_coords[vertex->texture_coord_id];

    /* draw 'em */
    draw_uv(wand, image, uv[0], uv[1]);
    draw_uv(wand, image, uv[1], uv[2]);
    draw_uv(wand, image, uv[2], uv[0]);
  }
  free(uv);

  /* write out the new image */
  if (snprintf(image->filename, MaxTextExtent, "%s.%u.%s", gsm->basename, set->triangle_group_id, ext) == -1) {
    ERROR("can't set output filename");
    DestroyImage(image);
    return 11;
  }
  MagickDrawImage(image_wand, wand);
  if (! MagickWriteImage(image_wand, image->filename)) {
    ERROR("failed to write output image!\n");
    DestroyImage(image);
    return 12;
  }

  if (filename) SAY("wrote UV maps for %s from %s to %s\n", material->name, filename, image->filename);
  else SAY("wrote UV maps for %s to %s\n", material->name, image->filename);

  DestroyImage(image);

  return 0;
}

int draw_uv(DrawingWand *wand, Image *image, struct texture_coord *uv0, struct texture_coord *uv1) {
  float u0, u1, v0, v1;
  float ut, vt; /* tmp */
  int x0, y0, x1, y1;

  /* backface culling :-) */
  if (uv0->u >= 0 && uv1->u < 0) return 1;
  if (uv0->v >= 0 && uv1->v < 0) return 2;

  /* absolute values */
  if (uv0->u < 0) u0 = 1.0 + uv0->u; else u0 = uv0->u;
  if (uv0->v < 0) v0 = 1.0 + uv0->v; else v0 = uv0->v;
  if (uv1->u < 0) u1 = 1.0 + uv1->u; else u1 = uv1->u;
  if (uv1->v < 0) v1 = 1.0 + uv1->v; else v1 = uv1->v;

  if (uv0->v < 0 && uv1->v >= 0) {
    /* wrap in V */
    if (uv0->u < 0 && uv1->u >= 0) {
      /* wrap in U */
      WARN("wrapping in both U and V space! please report BUGS!\n");
      vt = calculate_uv_wrap(u0, u1, v0, v1);
      ut = calculate_uv_wrap(v0, v1, u0, u1);
      x0 = (int) (u0 * image->columns); y0 = (int) (v0 * image->rows);
      x1 = (int) (ut * image->columns); y1 = (vt * image->rows);
      DrawLine(wand, x0, y0, x1, y1);
      x0 = 0; y0 = 0;
      x1 = (int) (u1 * image->columns); y1 = (v1 * image->rows);
      DrawLine(wand, x0, y0, x1, y1);
      return 0;
    }
    else {
      /* direct in U */
      ut = calculate_uv_wrap(v0, v1, u0, u1);
      x0 = (int) (u0 * image->columns); y0 = (int) (v0 * image->rows);
      x1 = (int) (ut * image->columns); y1 = image->rows;
      DrawLine(wand, x0, y0, x1, y1);
      x0 = (int) (ut * image->columns); y0 = 0;
      x1 = (int) (u1 * image->columns); y1 = (int) (v1 * image->rows);
      DrawLine(wand, x0, y0, x1, y1);
      return 0;
    }
  }
  else {
    /* direct in V */
    if (uv0->u < 0 && uv1->u >= 0) {
      /* wrap in U */
      vt = calculate_uv_wrap(u0, u1, v0, v1);
      x0 = (int) (u0 * image->columns); y0 = (int) (v0 * image->rows);
      x1 = image->columns; y1 = (int) (vt * image->rows);
      DrawLine(wand, x0, y0, x1, y1);
      x0 = 0; y0 = (int) (vt * image->rows);
      x1 = (int) (u1 * image->columns); y1 = (int) (v1 * image->rows);
      DrawLine(wand, x0, y0, x1, y1);
      return 0;
    }
    else {
      /* direct in U */
      x0 = (int) (u0 * image->columns); y0 = (int) (v0 * image->rows);
      x1 = (int) (u1 * image->columns); y1 = (int) (v1 * image->rows);
      DrawLine(wand, x0, y0, x1, y1);
      return 0;
    }
  }
}

/*
  find out where the line connecting two points over the edge should meet
  x is the one we're calculating; y is the other axis
*/
float calculate_uv_wrap(float x0, float x1, float y0, float y1) {
  float m, c;

  /* y = mx + c */
  m = (y1 - y0) / fabs(1.0 + x1 - x0);
  c = y0 - (m * x0);

  return m /* * {x = 1.0} */ + c;
}
#endif
