#include "gsmhacking.h"

mapping_value_t **alloc_mapping_values(unsigned short n) {
  return (mapping_value_t **) alloc_lots(n, "mapping_value", sizeof(mapping_value_t *), (void *(*)()) alloc_mapping_value);
}

mapping_value_t *alloc_mapping_value() {
  return (mapping_value_t *) alloc_one("mapping_value", sizeof(mapping_value_t));
}

mapping_t *alloc_mapping(unsigned short size) {
  mapping_t *ret;
 
  ret = (mapping_t *) alloc_one("mapping", sizeof(mapping_t));
  if (! ret) return 0;

  ret->pairs = alloc_mapping_values(size);
  if (! ret->pairs) {
    free(ret);
    return 0;
  }
#if 0
  ret->keys = (unsigned short **) alloc_lots(size, "mapping_key", sizeof(unsigned short *), (void *(*)()) alloc_short);
  if (! ret->keys) {
    free(ret);
    return 0;
  }
  ret->values = (unsigned short **) alloc_lots(size, "mapping_value", sizeof(unsigned short *), (void *(*)()) alloc_short);
  if (! ret->values) {
    free_structs("mapping_key", size, (void *) ret->keys);
    free(ret);
    return 0;
  }
#endif
  ret->size = size;

  return ret;
}

/* set all keys to their index, values to 0xffff => unused - hack alert!*/
void initialise_mapping(mapping_t *mapping) {
  unsigned short i;

  if (! mapping) {
    ERROR("no mapping!\n");
    return;
  }

  for (i = 0; i < mapping->size; i++) {
    mapping->pairs[i]->key = i;
    mapping->pairs[i]->value = 0xffff;
#if 0
    *mapping->keys[i] = i;
    *mapping->values[i] = 0xffff;
#endif
  }
}

void cleanup_mapping(mapping_t *mapping) {
  if (! mapping) {
    ERROR("no mapping!\n");
    return;
  }
  free_structs("mapping_value", mapping->size, (void **) mapping->pairs);
#if 0
  TRACE2("freeing keys at %p\n", mapping->keys);
  free_structs("mapping_key", mapping->size, (void *) mapping->keys);
  TRACE2("freeing values at %p\n", mapping->values);
  free_structs("mapping_value", mapping->size, (void *) mapping->values);
#endif
  TRACE2("freeing mapping at %p\n", mapping);
  free(mapping);
}

/* remove an element */
void pop_mapping(mapping_t *mapping, unsigned short index) {
  if (! mapping) {
    ERROR("no mapping!\n");
    return;
  }
  if (index >= mapping->size) {
    ERROR("index out of range\n");
    return;
  }

  free(mapping->pairs[index]);
  mapping->size--;
#if 0
  free(mapping->keys[index]);
  free(mapping->values[index]);
  mapping->size--;
#endif
}

/* -1 => not found, < -1 => error */
int get_value_for_mapping(mapping_t *mapping, unsigned short key) {
  unsigned short i;

  if (! mapping) {
    ERROR("no mapping\n");
    return -2;
  }

  for (i = 0; i < mapping->size; i++) {
    if (mapping->pairs[i]->key == key) return mapping->pairs[i]->value;
#if 0
    if (*mapping->keys[i] == key) return *mapping->values[i];
#endif
  }

  TRACE("key %u not found in mapping at %p\n", key, mapping);
  return -1;
}

/* -1 => not found, < -1 => error */
int get_key_for_mapping(mapping_t *mapping, unsigned short value) {
  unsigned short i;

  if (! mapping) {
    ERROR("no mapping\n");
    return -2;
  }

  for (i = 0; i < mapping->size; i++) {
    if (mapping->pairs[i]->value == value) return mapping->pairs[i]->key;
#if 0
    if (*mapping->values[i] == value) return *mapping->keys[i];
#endif
  }

  TRACE("value %u not found in mapping at %p\n", value, mapping);
  return -1;
}

/* -1 => not found, < -1 => error */
int set_value_for_mapping(mapping_t *mapping, unsigned short key, unsigned short value) {
  unsigned short i;

  if (! mapping) {
    ERROR("no mapping\n");
    return -2;
  }

  for (i = 0; i < mapping->size; i++) {
    if (mapping->pairs[i]->key != key) continue;
    mapping->pairs[i]->value = value;
#if 0
    if (*mapping->keys[i] != key) continue;
    *mapping->values[i] = value;
#endif
    return 0;
  }

  ERROR("key %u not found in mapping at %p\n", key, mapping);
  return -1;
}

/* -1 => not found, < -1 => error */
int set_key_for_mapping(mapping_t *mapping, unsigned short value, unsigned short key) {
  unsigned short i;

  if (! mapping) {
    ERROR("no mapping\n");
    return -2;
  }

  for (i = 0; i < mapping->size; i++) {
    if (mapping->pairs[i]->value != value) continue;
    mapping->pairs[i]->key = key;
    return 0;
  }

  ERROR("key %u not found in mapping at %p\n", key, mapping);
  return -1;
}

int sort_mapping_by_value(mapping_t *mapping, int normalise) {
  unsigned short i;

  if (! mapping) {
    ERROR("no mapping!\n");
    return 1;
  }

  /* sort by value */
  qsort((void *) mapping->pairs, (size_t) mapping->size, sizeof(mapping_value_t *), sort_mapping_by_value_function);
  /* and reorder keys */
  if (normalise) for (i = 0; i < mapping->size; i++) mapping->pairs[i]->key = i;

  return 0;
}

/* 0 on success */
int set_key_and_value_for_mapping(mapping_t *mapping, unsigned short index, unsigned short key, unsigned short value) {
  if (! mapping) {
    ERROR("no mapping\n");
    return -1;
  }

  if (index >= mapping->size) {
    ERROR("index out of range\n");
    return -2;
  }

  mapping->pairs[index]->key = key;
  mapping->pairs[index]->value = value;

  return 0;
}

/* <0 on error */
int get_key_in_mapping(mapping_t *mapping, unsigned short index) {
  if (! mapping) {
    ERROR("no mapping\n");
    return -1;
  }

  if (index >= mapping->size) {
    ERROR("index out of range\n");
    return -2;
  }

  return mapping->pairs[index]->key;
}

/* <0 on error */
int get_value_in_mapping(mapping_t *mapping, unsigned short index) {
  if (! mapping) {
    ERROR("no mapping\n");
    return -1;
  }

  if (index >= mapping->size) {
    ERROR("index out of range\n");
    return -2;
  }

  return mapping->pairs[index]->value;
}

int sort_mapping_by_value_function(const void *a, const void *b) {
  mapping_value_t *one, *two;

  if (! a || ! b) {
    ERROR("null argument(s)\n");
    return 0;
  }

  one = *(mapping_value_t **) a;
  two = *(mapping_value_t **) b;

  TRACE("one: %p <%u, %u>\n", one, one->key, one->value);
  TRACE("two: %p <%u, %u>\n", two, two->key, two->value);

  if (one->value < two->value) return -1;
  if (one->value > two->value) return 1;

  return 0;
}

void debug_mapping(mapping_t *mapping, char *thing) {
  unsigned short i;

  if (! gsm_debug) return;

  if (! mapping) {
    ERROR("no mapping!\n");
    return;
  }

  DEBUG("%s mapping has %u element%s\n", thing, mapping->size, (mapping->size == 1) ? "" : "s");
  for (i = 0; i < mapping->size; i++) {
    DEBUG("map %u: %s %u -> %u\n", i, thing, mapping->pairs[i]->key, mapping->pairs[i]->value);
#if 0
    DEBUG("map %u: %s %u -> %u\n", i, thing, *mapping->keys[i], *mapping->values[i]);
#endif
  }
}
