#include "gsmhacking.h"

void **alloc_lots(unsigned short n, char *thing, size_t size_lots, void *(*alloc_function)()) {
  unsigned short i;
  void **ret;
  
  /* allocate the pointer plus a null */
  ret = malloc(n * size_lots);
  TRACE2("alloc_%ss(%u * %u): %x\n", thing, size_lots, n, ret);
  if (! ret) {
    ERROR("out of memory for **%s", thing);
    return 0;
  }
 
  /* allocate the members only if an allocation function was provided */
  if (alloc_function) {
    for (i = 0; i < n; i++) {
      ret[i] = alloc_function();
      if (! ret[i]) return 0;
    }
  }

  return ret;
}

/* should have done this before: "increase" an array size */
void **alloc_more(unsigned short n, char *thing, size_t size_lots, void *(*alloc_function)(), unsigned short num_src, void **src, size_t size_one) {
  void **dest;
  unsigned short i;

  dest = malloc((n + num_src) * size_lots);
  TRACE2("alloc_%ss(%u * %u): 0x%x\n", thing, size_lots, n + num_src, dest);
  if (! dest) {
    ERROR("out of memory for new %s array\n", thing);
    return 0;
  }

  /* allocate the members */
  for (i = 0; i < n + num_src; i++) {
    dest[i] = alloc_function();
    if (! dest[i]) return 0;
    if (i < num_src) memmove(dest[i], src[i], size_one);
  }
  free_structs(thing, num_src, src);

  return dest;
}

/* generic allocation function */
void *alloc_one(char *thing, size_t size) {
  void *ret;

  ret = malloc(size);
  TRACE2("alloc_%s(%u): 0x%x\n", thing, size, ret);
  if (! ret) ERROR("out of memory for *%s", thing);
  memset(ret, 0, size);

  return ret;
}

/* allocate enough memory for a name and write into it */
unsigned char *alloc_name(char *name, unsigned short *dest_len) {
  unsigned char *dest;
  int name_length;
 
  /* first set dest_len to zero so we won't try to free() it later */
  *dest_len = 0;

  /* allocate enough space */
  name_length = strlen(name);
  if (name_length > 0xffff) {
    ERROR("string name too long, max length %u characters\n", 0xffff);
    return 0;
  }
  dest = (unsigned char *) malloc (sizeof(unsigned char) * (name_length + 1));
  TRACE2("malloc(%u * %d): %x\n", sizeof(unsigned char), name_length + 1, dest);
  if (! dest) {
    ERROR("out of memory for name");
    return 0;
  }
  TRACE2("memmove({dest=%p}, {\"%s\"=%p}, {name_length=%d})\n", dest, name, name, name_length);
  memmove(dest, name, name_length);

  /* clean up */
  dest[name_length] = '\0';
  *dest_len = name_length;

  return dest;
}

unsigned char *alloc_char() {
  return (unsigned char *) alloc_one("char", sizeof(unsigned char));
}

unsigned short *alloc_short() {
  return (unsigned short *) alloc_one("short", sizeof(unsigned short));
}

float *alloc_float() {
  return (float *) alloc_one("float", sizeof(float));
}

/* walk through the array and free() everything */
void free_structs(char *name, unsigned short num_things, void **thing) {
  int i;

  if (num_things && ! thing) {
    ERROR("SEVERE BUG: num_%ss=%u, **%s is null\n", name, num_things, name);
    return;
  }

  for (i = 0; i < num_things; i++) {
    if (thing[i]) {
      TRACE2("freeing %s at 0x%x\n", name, thing[i]);
      free(thing[i]);
    }
    else ERROR("null %s\n", name);
  }

  if (num_things) {
    if (thing) {
      TRACE2("freeing pointer at 0x%x\n", thing);
      free(thing);
    }
    else ERROR("null %s pointer\n", name);
  }
}

/* print data in a format that can be copy and pasted into C code */
void format_unknown(char *name, unsigned char *src, unsigned int length) {
  unsigned int i;
  char *unknown;

  unknown = (char *) malloc((length * 4) + 1);
  if (! unknown) {
    ERROR("out of memory for unknown formatting");
    return;
  }

  for (i = 0; i < length; i++) {
    snprintf(unknown + (i * 4), 5, "\\x%02x", src[i]);
  }
  unknown[length * 4] = '\0';
  TRACE1("%s unknown data: %s\n", name, unknown);
  free(unknown);
}

int str_equals(char *one, char *two) {
  int i;
  for (i = 0; ; i++) {
    if (two[i] != one[i]) return 0;
    if (! one[i]) return 1;
  }
}

int str_equiv(char *one, char *two) {
  int i;
  for (i = 0; ; i++) {
    if (tolower(two[i]) != tolower(one[i])) return 0;
    if (! one[i]) return 1;
  }
}

void catch_segv() {
  error_dbg("\n");
  error_dbg("\t\t\t           +--------+\n");
  error_dbg("\t\t\t         _-+  SHIT! |\n");
  error_dbg("\t\t\t  |\\_/| -  +--------+\n");
  error_dbg("\t\t\t =(o.o)=\n");
  error_dbg("\t\t\t  (   ),\n");
  error_dbg("\t\t\t  ^^ ^^              \n");
  error_dbg("\n");
  error_dbg("gsmhacking segfaulted, which means it tried to do something very bad!\n");
  error_dbg("if you compiled gsmhacking from source, please make clean and recompile\n");
  error_dbg("if you don't have the source or if you still get this message after\n");
  error_dbg("recompiling, please send the full output of\n\n");
  error_dbg("    gsmhacking --super-debug <your options>\n\n");
  error_dbg("where <your options> are the exact arguments you gave to gsmhacking,\n");
  error_dbg("along with the GSM file you were working on, to gsmhacking@furrycat.net\n");
  exit(100);
}

char *get_dirname(char *filename, unsigned short *position, int *need_free) {
  char *ret, *s;

  *need_free = 0;

  if (! filename || ! *filename) {
    ERROR("null filename\n");
    return 0;
  }

  /* optimisation: no directory separator => return the filename */
  for (s = filename; *s; s++) if (*s == PATHSEP) break;
  if (! *s) {
    *position = 0;
    return "";
  }

  ret = (char *) alloc_name(filename, position);
  if (! position) {
    ERROR("out of memory for dirname");
    return 0;
  }
  *need_free = 1;

  for (s = ret; *s; s++);
  for (s-- ; *s; s--) {
    if (*s == PATHSEP) {
      *s = '\0';
      break;
    }
    --*position;
  }
  DEBUG("ret: %s\n", ret);

  return ret;
}

char *add_suffix(char *filename, char *suffix) {
  char *ret;
  int len;

  if (! filename) {
    ERROR("no filename to suffix\n");
    return 0;
  }

  DEBUG("filename: %s\n", filename);
  len = strlen(filename) + strlen(suffix) + 1;
  ret = (char *) malloc(len);
  if (! ret) {
    ERROR("out of memory for suffix");
    return 0;
  }

  if (snprintf(ret, len, "%s%s", filename, suffix) == -1) {
    ERROR("failed to add suffix");
    return 0;
  }
  DEBUG("ret: %s\n", ret);

  return ret;
}

char *add_prefix(char *filename, char *prefix) {
  char *ret, *s;
  int prunt, len, need_free;
  unsigned short i;

  if (! filename) {
    ERROR("no filename to prefix\n");
    return 0;
  }

  DEBUG("filename: %s\n", filename);
  len = strlen(filename) + strlen(prefix) + 1;
  ret = (char *) malloc(len);
  if (! ret) {
    ERROR("out of memory for prefix");
    return 0;
  }

  /* find dir separator */
  s = get_dirname(filename, &i, &need_free);
  DEBUG("%s, %u\n", s, i);

  /* write dirname */
  prunt = snprintf(ret, len, "%s", s);
  if (need_free) free(s);
  if (prunt == -1) {
    ERROR("failed to write dirname");
    return 0;
  }
  DEBUG("ret: %s\n", ret);

  /* replace dir separator */
#ifdef WIN32
  *(ret + i - 1) = '\\';
#else
  *(ret + i - 1) = '/';
#endif

  /* stick prefix in front */
  prunt = snprintf(ret + i, len, "%s", prefix);
  if (prunt == -1) {
    ERROR("failed to write prefix");
    return 0;
  }
  DEBUG("ret: %s\n", ret);

  /* write rest of string */
  prunt = snprintf(ret + i + prunt, len, "%s", filename + i);
  if (prunt == -1) {
    ERROR("failed to complete string");
    return 0;
  }
  DEBUG("ret: %s\n", ret);

  return ret;
}
