#include "gsmhacking.h"

extern char **filenames;
struct option **alloc_options(unsigned short n) {
  return (struct option **) alloc_lots(n, "option", sizeof(struct option *), (void *(*)()) alloc_option);
}

struct option *alloc_option() {
  struct option *ret;

  ret = (struct option *) alloc_one("option", sizeof(struct option));
  if (! ret) return 0;
  initialise_option(ret);

  return ret;
}

void cleanup_option(struct option *option) {
  if (! option) {
    DEBUG("null option pointer (might be because we were called from a transplant\n");
    return;
  }
  if (--option->gsmcount) {
    DEBUG("option at %p still in use; gsmcount: %u\n", option, option->gsmcount);
    return;
  }
  if (option->texture_searches) {
    TRACE2("freeing texture_searches option at %p\n", option->texture_searches);
    cleanup_texture_searches(option->texture_searches);
  }
  if (option->texture_replaces) {
    DEBUG("option->texture_replaces: %p next: %p\n", option->texture_replaces, option->texture_replaces);
  }
  if (option->texture_sets) {
    DEBUG("option->texture_sets: %p next: %p\n", option->texture_sets, option->texture_sets->next);
  }
  if (option->map_sets) {
    DEBUG("option->map_sets: %p next: %p\n", option->map_sets, option->map_sets->next);
  }
  if (option->material_sets) {
    DEBUG("option->material_sets: %p next: %p\n", option->material_sets, option->material_sets->next);
  }
  if (option->outfile) {
    DEBUG("option->outfile: %s\n", option->outfile);
  }
  if (option->prefix) {
    DEBUG("option->prefix: %s\n", option->prefix);
  }
  if (option->suffix) {
    DEBUG("option->suffix: %s\n", option->suffix);
  }
  if (option->hit_detect) {
    DEBUG("option->hit_detect: %s\n", option->hit_detect);
  }
  if (option->event_sets) {
    DEBUG("option->event_sets: %p next: %p\n", option->event_sets, option->event_sets->next);
  }
  if (option->transplant_sets) {
    DEBUG("option->transplant_sets: %p next: %p\n", option->transplant_sets, option->transplant_sets->next);
  }
  if (option->copy_material_sets) {
    DEBUG("option->copy_material_sets: %p next: %p\n", option->copy_material_sets, option->copy_material_sets->next);
  }
  if (option->rotation_sets) {
    DEBUG("option->rotation_sets: %p next: %p\n", option->rotation_sets, option->rotation_sets->next);
  }
#ifdef USE_IMAGEMAGICK
  if (option->generate_uv_map_sets) {
    DEBUG("option->generate_uv_map_sets: %p next: %p\n", option->generate_uv_map_sets, option->generate_uv_map_sets->next);
    //cleanup_generate_uv_map_sets(option->generate_uv_map_sets);
  }
#endif
  TRACE2("freeing option at %p\n", option);
  free(option);
}

void initialise_option(struct option *option) {
  memset(option, 0, sizeof(struct option));
}

void cleanup_options(unsigned short num_options, struct option **options) {
  free_structs("option", num_options, (void *) options);
}

#define skip(x) \
  { for (j = i; j < argc - x; j++) argv[j] = argv[j + x]; argc -= x; i--; }
struct gsm **process_options(int *_argc, char **_argv, int *num_gsms) {
  struct gsm **gsms;
  struct option **options, option;
  int i, j, num_skip;
  unsigned short len;
  int cur;
  int argc;
  char **argv;
  int last_option;
  int num_filegroups;

  *num_gsms = num_filegroups = 0;

  /* must have at least one gsm */
  argc = *_argc - 1;
  if (! argc) {
    usage();
    return 0;
  }
  argv = _argv + 1;

  /* worst case scenario is all arguments are option groups */
  options = alloc_options(argc);
  if (! options) {
    ERROR("out of memory for options");
    *num_gsms = -1;
    return 0;
  }

  /* or, worst case, all are files */
  gsms = alloc_gsms(argc);
  if (! gsms) {
    ERROR("out of memory for gsms");
    *num_gsms = -1;
    return 0;
  }

  cur = 0;
  last_option = OPTION_NONE;
  for (i = 0; i < argc; i++) {
    initialise_option(&option);
    num_skip = parse_option(&option, argc, argv, i);
    if (num_skip < 0) {
      for (i = 0; i < *_argc - 1; i++) {
        cleanup_option(options[i]);
        cleanup_gsm(gsms[i]);
      }
      free(options);
      free(gsms);
      if (num_skip % 2) (void) usage();
      return 0;
    }
    if (num_skip) {
      /* argument */
      if (last_option == OPTION_FILENAME && options[cur]->number) {
        cur++;
        DEBUG("new option group with \"%s\": %d\n", argv[i], cur);
      }
      last_option = OPTION_ARGUMENT;
      copy_options(options[cur], &option);
      if (num_skip) skip(num_skip);
    }
    else {
      /* filename */
      DEBUG("option not recognised, treating as file: %s\n", argv[i]);
      gsms[*num_gsms]->filename = alloc_name(argv[i], &len);
      if (! len) {
        ERROR("out of memory for file %d name\n", *num_gsms);
        *num_gsms = -1;
        return 0;
      }
      gsms[*num_gsms]->option = options[cur];
      gsms[*num_gsms]->option->gsmcount++;
      DEBUG("gsm %d filename %s at %p has option %d at %p (gsmcount=%u)\n", *num_gsms, gsms[*num_gsms]->filename, gsms[*num_gsms]->filename, cur, options[cur], gsms[*num_gsms]->option->gsmcount);
      ++*num_gsms;
      if (last_option != OPTION_FILENAME) num_filegroups++;

      last_option = OPTION_FILENAME;
    }
  }

  /* this isn't BSD */
  DEBUG("current option: %d, num_filegroups: %d\n", cur, num_filegroups);
  if (num_filegroups && cur >= num_filegroups) {
    DEBUG("handling trailing options (cur: %d, num_filegroups: %d)\n", cur, num_filegroups);
    for (i = num_filegroups; i <= cur; i++) {
      DEBUG("appending option group %d to group %d\n", i, num_filegroups - 1);
      copy_options(options[num_filegroups - 1], options[i]);
    }
  }

  /* free unused options and gsms */
  argc = *_argc - 1;
  if (argc == *num_gsms) cur = 1;
  else cur = argc - *num_gsms;
  DEBUG("cur: %d, *num_gsms: %d, num_filegroups: %d, argc: %d\n", cur, *num_gsms, num_filegroups, argc);
  for (i = cur; i < argc; i++) {
    if (options[i]->number) {
      ERROR("please tell furrycat he's a tosser (%d)\n", i);
    }
    DEBUG("freeing unused option %d at %p with number %u\n", i, options[i], options[i]->number);
    free(options[i]);
  }
  free(options);

  for (i = *num_gsms; i < argc; i++) {
    if (gsms[i]->filename) {
      ERROR("please tell furrycat he's a wanker (%d)\n", i);
    }
    DEBUG("freeing unused GSM %d at %p\n", i, gsms[i]);
    free(gsms[i]);
  }

  /* no filenames => bad */
  if (! num_filegroups) {
    (void) usage();
    return 0;
  }

  return gsms;
}

#if 0
/* given an offset list like 1234,5678 we parse the offsets */
unsigned int **parse_offsets(char *arg) {
  char *s, *t, *bogus;
  int i, j;
  unsigned int number, **ret;

  /* first count up how many items there are in the list   */
  /* start from 1 because there's at least one in the list */
  for (i = 1, s = arg; *s; s++) if (*s == ',') i++;

  /* now set up the array (or "unsigned int star star" if you prefer) */
  ret = (unsigned int **) malloc(sizeof(unsigned int *) * (i + 1));
  TRACE("parse_offsets(): malloc({(sizeof(unsigned int *) * %d)=%d}): %p\n", i + 1, (sizeof(unsigned int *) * (i + 1)), ret);
  if (! ret) {
    ERROR("oops, out of memory for offsets list\n");
    exit(102);
  }
  ret[i] = (unsigned int *) 0;

  i = 0;
  for (s = arg; *s; s++) {
    if (*s < '0' || *s > '9')
      if (*s < 'a' || *s > 'f')
        if (*s < 'A' || *s > 'F')
          if (*s != 'x' || *s != 'X')
            if (*s != ',') goto bad_offsets;
    /* oops, that'll be the end of the list then */
    if (! *s) break;
    for (t = s; *t && *t != ','; t++);
    /* found a delimiter */
    if (*t == ',') {
      *t = '\0';
      TRACE("parse_offsets(): %s!\n", s);
    }
    else t--;

    /* first parse the number */
    number = strtol(s, &bogus, 16);
    TRACE("parse_offsets(): strtol(\"%s\", %p, 16): 0x%04x\n", s, bogus, number);
    if (*bogus) goto bad_offsets;

    /* check this offset wasn't set already */
    for (j = 0; ret[j]; j++) {
      if (*(ret[j]) == number) {
        DEBUG("parse_offsets(): 0x%04x already in the list as ret[%d] at %p\n", number, j, ret[j]);
        goto moving_on;
      }
    }
    /* allocate memory for it */
    ret[i] = (unsigned int *) malloc(sizeof(unsigned int));
    TRACE("parse_offsets(): malloc({(sizeof(unsigned int))=%d}): {ret[%d]=%p}\n", sizeof(unsigned int), i, ret[i]);
    if (! ret[i]) {
      ERROR("oops, out of memory for offsets list\n");
      exit(102);
    }
    /* and fill it in */
    *(ret[i]) = number;
    DEBUG("parse_offsets(): adding 0x%04x as ret[%d] at %p\n", number, i, ret[i]);
    i++;

moving_on:
    /* move to next item */
    s = t;
  }

  DEBUG("parse_offsets(): returning %p\n", ret);
  return ret;

bad_offsets:
  TRACE("\n");
  ERROR("invalid offset list - try 0x1234,0x5678 etc\n");
  DEBUG("parse_offsets(): freeing ret and bailing out\n");
  for (j = 0; j < i; j++) free(ret[j]);
  free(ret);
  exit(107);
}
#endif

int usage() {
  error_dbg("usage: gsmhacking [options] filenames [[options] [filenames]]\n");
  error_dbg("options: -v,  --verbose                 [be verbose]\n");
#ifdef EBUG
  error_dbg("         -d,  --debug                   [debugging mode]\n");
  error_dbg("         -sd, --super-debug             [hardcore debugging mode]\n");
#endif
  error_dbg("         -q,  --quiet                   [no output at all]\n");
  error_dbg("         -c,  --classify                [show type of GSM]\n");
  error_dbg("         -S,  --summary                 [print summary of GSM properties]\n");
  error_dbg("         -D   --describe                [show detailed description of GSM]\n");
  error_dbg("         -l,  --list-textures           [list textures in .gsm]\n");
  error_dbg("         -s,  --texture-search filename [search for a filename]\n");
  error_dbg("         -r,  --texture-replace old new [replace old texture with new]\n");
  error_dbg("         -t,  --texture-set number file [assign filename to numbered texture]\n");
  error_dbg("         -mt, --map-texture material type number\n");
  error_dbg("                                        [assign texture number to material]\n");
  error_dbg("         -T,  --transparency number value\n");
  error_dbg("                                        [set material number transparency]\n");
  error_dbg("         -cm, --clear-map material      [remove all maps from a material]\n");
  error_dbg("         -N,  --name\n");
  error_dbg("                                        [set a material's name]\n");
  error_dbg("         -F,  --add-flashlight          [add a flashlight to weapon model]\n");
  error_dbg("         -R,  --make-rifle              [convert pistol to rifle]\n");
  error_dbg("              --make-swt                [convert model to good guy]\n");
  error_dbg("              --make-kid                [convert model to kid]\n");
  error_dbg("              --make-tac                [convert model to bad guy]\n");
  error_dbg("              --make-bf1                [convert model to bad girl]\n");
  error_dbg("              --default-skeleton        [reset model's bind pose]\n");
  error_dbg("              --copy-material file material name\n");
  error_dbg("                                        [copy material groups from file]\n");
  error_dbg("              --transplant file bone type\n");
  error_dbg("                                        [transplant bone from file]\n");
  error_dbg("              --rotate-bone bone axis angle start end\n");
  error_dbg("                                        [rotate bone angle degrees]\n");
  error_dbg("                                        [about the given axis in frames]\n");
  error_dbg("                                        [start through end]\n");
#ifdef USE_IMAGEMAGICK
  error_dbg("         -uv, --generate-uv-maps group colour\n");
  error_dbg("                                        [make a UV map file for given group]\n");
#endif
  error_dbg("              --optimise-materials      [remove redundant materials]\n");
  error_dbg("              --optimise-textures       [remove redundant textures]\n");
  error_dbg("         -O,  --optimise                [perform all possible optimisations]\n");
  error_dbg("         -o,  --output filename         [write new file]\n");
  error_dbg("              --outfile filename\n");
  error_dbg("              --prefix prefix\n");
  error_dbg("                                        [add prefix to filename for output]\n");
  error_dbg("              --suffix suffix\n");
  error_dbg("                                        [add suffix to filename for output]\n");
  error_dbg("         -p,  --polycount               [show mesh polygon (triangle) count]\n");
  error_dbg("              --list-map-types          [list recognisded map types]\n");
  error_dbg("              --export-skin-o-matic     [write data for skin-o-matic to parse]\n");
  error_dbg("         -E   --export-milkshape        [write Milkshape 3D file]\n");
  error_dbg("         -I   --import-milkshape        [read Milkshape 3D file]\n");
  error_dbg("              --john-woo                [slow down an animation]\n");
  error_dbg("         -ce  --clear-events frame      [clear events from a frame]\n");
  error_dbg("         -ae  --add-event frame event   [add an event to a frame]\n");
  error_dbg("         -fh  --fix-hit-detect file     [import hit detection information]\n");
  error_dbg("         -h,  --help                    [this information]\n");
  error_dbg("         -V,  --version                 [this information]\n");
  error_dbg("notes: --summary is implicit when no options are given\n");
  error_dbg("       all options preceding filenames are applied to those files\n");
  error_dbg("       options following files are applied to the next set of files\n");
  error_dbg("       use \"--outfile -\" to write new GSM to stdout\n");
  error_dbg("       when using --map-texture you may specify the material by name or\n");
  error_dbg("       number and the map types by name or number (use --list-map-types for\n");
  error_dbg("       a list of recognised map types\n");
  error_dbg("       transparency value is an integer percentage\n");
  error_dbg("       material options can be called multiple times for one GSM\n");
  error_dbg("       use -d and -v together to get more debugging info\n");
  error_dbg("       if more than one of -d/-v or -q is given, the last takes precedence\n");
  return 1;
}

int help() {
  usage();
  error_dbg("examples: gsmhacking --classify BG1_VIC_GET_HIGH.GSM\n");
  error_dbg("          [determine type of Victor Getts GSM file (should print TAC)]\n\n");
  error_dbg("          gsmhacking --classify SWT_ELEM_HIGH.GSM --summary M4.GSM\n");
  error_dbg("          [determine type of SWT_ELEM_HIGH and show statistics for M4]\n\n");
  error_dbg("          gsmhacking --make-swt PRH_HIS_HOL.GSM --outfile SWT_VICAR.GSM\n");
  error_dbg("          [convert His Holiness to SWAT officer in file SWT_VICAR.GSM]\n\n");
  error_dbg("          gsmhacking --make-tac SWT_*.GSM --classify SWT_ELEM_HIGH.GSM\n");
  error_dbg("          [convert all SWT models to bad guys then classify SWT_ELEM_HIGH]\n\n");
  error_dbg("          gsmhacking *.GSM --add-flashlight\n");
  error_dbg("          [add a flashlight to all GSM files]\n\n");
  error_dbg("          gsmhacking SWT_BEEF_HIGH.GSM --map-texture HELMET diffuse 1\n");
  error_dbg("          [set \"HELMET\" material to a diffuse map of texture #01]\n\n");
  error_dbg("          gsmhacking SWT*.GSA --make-tac --prefix TAC_\n");
  error_dbg("          [convert all SWT animations to TAC writing, eg TAC_SWT_blabla.GSA]\n\n");
  return 1;
}

/* macro to remove processed arguments from the argv list */
/* check if argv[i] is valid and put it in option if necessary */
#define argument(x) str_equals(argv[i], x)
int parse_option(struct option *option, int argc, char **argv, int i) {
  unsigned short len;
  char *search, *replace;

  if (argument("-V") || argument("--version")) {
    exit(version());
  }
  else if (argument("-d") || argument("--debug")) {
    gsm_debug = 1; gsm_quiet = 0;
    gsm_error = stdout;
    DEBUG("entering DEBUG mode\n");
    return 1;
  }
  else if (argument("-h") || argument("--help")) {
    exit(help());
  }
  else if (argument("--list-map-types")) {
    exit(list_map_types());
  }
#if 0
  else if (argument("-O") || argument("--offset")) {
    if (i == *argc - 1) return 0;
    offsets = parse_offsets(argv[i + 1]);
    DEBUG("offsets array is at %p\n", offsets);
    if (! offsets) return 0;
    skip(2);
  }
#endif
  else if (argument("-l") || argument("--list-textures") || argument("--list-texture") || argument("--texture-list") || argument("--textures-list")) {
    option->texture_list = 1;
    DEBUG("option --list-textures\n");
    option->number++;
    return 1;
  }
  else if (argument("-m") || argument("--missing")) {
    option->showmissing = 1;
    DEBUG("will report any files that are missing\n");
    return 1;
  }
  else if (argument("-o") || argument("--output") || argument("--outfile")) {
    if (i >= argc - 1) return 0;
    if (str_equals(argv[i + 1], "-")) {
      option->fd = stdout;
      gsm_quiet = 1;
      return 2;
    }
    else if (option->fd) {
      error_dbg("can't mix --outfile <file> and --outfile -\n");
      return -1;
    }
    option->outfile = alloc_name(argv[i + 1], &len);
    if (! len) {
      ERROR("out of memory for output filename");
      return -1;
    }
    DEBUG("output file is \"%s\"\n", option->outfile);
    return 2;
  }
  else if (argument("-p") || argument("--polycount") || argument("--polygon-count")) {
    DEBUG("will show this model's polygon count\n");
    option->polycount = 1;
    option->number++;
    return 1;
  }
  else if (argument("-q") || argument("--quiet")) {
    gsm_quiet = 1; gsm_verbose = gsm_debug = 0;
    DEBUG("entering QUIET mode - if you can read this it's a bug!\n");
    return 1;
  }
  else if (argument("-r") || argument("--texture-replace") || argument("--replace-texture") || argument("--replace-textures")) {
    if (i >= argc - 2) return -1;
    search = argv[i + 1];
    replace = argv[i + 2];
    DEBUG("search string is \"%s\" and replacement string is \"%s\"\n", search, replace);
    if (add_to_texture_replace(&option->texture_replaces, search, replace)) {
      ERROR("can't add texture strings to search and replace list");
      return -1;
    }
    option->number++;
    return 3;
  }
  else if (argument("-s") || argument("--texture-search") || argument("--search-texture") || argument("--search-textures")) {
    if (i >= argc - 1) return -1;
    search = argv[i + 1];
    DEBUG("search string is \"%s\"\n", search);
    if (add_to_texture_search(&option->texture_searches, search)) {
      ERROR("can't add texture string to search list\n");
      return -1;
    }
    option->number++;
    return 2;
  }
  else if (argument("-t") || argument("--texture-set") || argument("--set-texture")) {
    if (i >= argc - 2) return -1;
    search = argv[i + 1];
    /* variable re-use */
    replace = argv[i + 2];
    DEBUG("set texture %s to \"%s\"\n", search, replace);
    if (add_to_texture_set(&option->texture_sets, search, replace)) {
      ERROR("can't add to texture set list\n");
      return -1;
    }
    option->number++;
    return 3;
  }
  else if (argument("-mt") || argument("--map-texture") || argument("--material-texture") || argument("--set-map-texture") || argument("--set-material-texture")) {
    if (i >= argc - 3) return -1;
    DEBUG("setting map %s to texture %s\n", argv[i + 1], argv[i + 2]);
    if (add_to_map_set(&option->map_sets, argv[i + 1], argv[i + 2], argv[i + 3])) {
      ERROR("can't set map\n");
      return -1;
    }
    option->number++;
    return 4;
  }
  else if (argument("-cm") || argument("--clear-map") || argument("--clear-material")) {
    if (i >= argc - 1) return -1;
    DEBUG("clearing map list from material %s\n", argv[i + 1]);
    if (add_to_map_set(&option->map_sets, argv[i + 1], 0, 0)) {
      ERROR("can't set map for clearing\n");
      return -1;
    }
    option->number++;
    return 2;
  }
  else if (argument("-T") || argument("--transparency") || argument("--material-transparency")) {
    if (i >= argc - 2) return -1;
    DEBUG("setting material %s transparency\n", argv[i + 1]);
    if (add_to_material_set(&option->material_sets, MATERIAL_PROPERTY_TRANSPARENCY, argv[i + 1], argv[i + 2], 0, 0)) {
      ERROR("can't set material set\n");
      return -1;
    }
    option->number++;
    return 3;
  }
  else if (argument("-N") || argument("--name") || argument("--material-name")) {
    if (i >= argc - 2) return -1;
    DEBUG("setting material %s's name to %s\n", argv[i + 1], argv[i + 2]);
    if (add_to_material_set(&option->material_sets, MATERIAL_PROPERTY_NAME, argv[i + 1], argv[i + 2], 0, 0)) {
      ERROR("can't set material set\n");
      return -1;
    }
    option->number++;
    return 3;
  }
  else if (argument("-v") || argument("--verbose")) {
    gsm_verbose++; gsm_quiet = 0;
    DEBUG("entering TRACE mode\n");
    return 1;
  }
  else if (argument("-sd") || argument("--super-debug")) {
    gsm_debug = 1; gsm_verbose = 999; gsm_quiet = 0;
    gsm_error = stdout;
    DEBUG("entering extremely verbose debugging mode\n");
    return 1;
  }
  else if (argument("--make-swt") || argument("--make-swat")) {
    option->convert = GSM_TYPE_SWT;
    DEBUG("will try to make this into a SWT model\n");
    option->number++;
    return 1;
  }
  else if (argument("--make-kid")) {
    option->convert = GSM_TYPE_KID;
    DEBUG("will try to make this into a KID model\n");
    option->number++;
    return 1;
  }
  else if (argument("--make-sus") || argument("--make-tac") || argument("--make-suspect")) {
    option->convert = GSM_TYPE_TAC;
    DEBUG("will try to make this into a TAC model\n");
    option->number++;
    return 1;
  }
  else if (argument("--make-bf1") || argument("--make-female-sus") || argument("--make-female-suspect")) {
    option->convert = GSM_TYPE_BF1;
    DEBUG("will try to make this into a BF1 animation\n");
    option->number++;
    return 1;
  }
  else if (argument("--default-skeleton")) {
    option->convert = GSM_TYPE_DEFAULT;
    DEBUG("will try to reset default skeleton\n");
    option->number++;
    return 1;
  }
  else if (argument("-R") || argument("--make-rifle")) {
    option->make_rifle = 1;
    DEBUG("will try to make this into rifle\n");
    option->number++;
    return 1;
  }
  /* XXX: bin me */
  else if (argument("--make-stl")) {
    option->make_stl = 1;
    SAY("XXX: will try to generate a new shared triangle list\n");
    option->number++;
    return 1;
  }
  else if (argument("-F") || argument("--add-flashlight")) {
    option->add_flashlight = 1;
    DEBUG("will try to add a flashlight\n");
    option->number++;
    return 1;
  }
  else if (argument("-D") || argument("--describe")) {
    if (! option->describe) option->describe = DESCRIBE_STANDARD;
    DEBUG("will dump extended information about this GSM\n");
    option->number++;
    return 1;
  }
  else if (argument("-O"), argument("--optimise") || argument("--optimize")) {
    option->optimise = 1;
    DEBUG("will try to optimise this GSM\n");
    option->number++;
    return 1;
  }
  else if (argument("--optimise-materials") || argument("--optimize-materials")) {
    option->optimise_materials = 1;
    DEBUG("will try to fix redundant materials\n");
    option->number++;
    return 1;
  }
  else if (argument("--optimise-textures") || argument("--optimize-textures")) {
    option->optimise_textures = 1;
    DEBUG("will try to fix redundant textures\n");
    option->number++;
    return 1;
  }
  else if (argument("-S") || argument("--summarise") || argument("--summarize") || argument("--summary")) {
    option->summary = 1;
    DEBUG("will dump summary information about this GSM\n");
    option->number++;
    return 1;
  }
  else if (argument("-c") || argument("--classify")) {
    option->classify = 1;
    DEBUG("will classify this GSM\n");
    option->number++;
    return 1;
  }
  else if (argument("--export-skin-o-matic")) {
    option->describe = DESCRIBE_SKIN_O_MATIC;
    DEBUG("will print skin-o-matic information\n");
    option->number++;
    return 1;
  }
  else if (argument("-E") || argument("--export-milkshape")) {
    option->export = EXPORT_MILKSHAPE;
    DEBUG("will write a Milkshape 3D representation of this GSM\n");
    option->number++;
    return 1;
  }
  else if (argument("-I") || argument("--import-milkshape")) {
    option->export = IMPORT_MILKSHAPE;
    DEBUG("will read a Milkshape 3D file and construct GSM\n");
    option->number++;
    return 1;
  }
  else if (argument("--john-woo") || argument("--johnwoo")) {
    option->frame_repeat_every = 1;
    option->frame_repeat_count = 2;
    DEBUG("slowing down this animation\n");
    option->number++;
    return 1;
  }
  else if (argument("--prefix")) {
    if (i >= argc - 1) return -1;
    option->prefix = argv[i + 1];
    DEBUG("file prefix: %s\n", option->prefix);
    option->number++;
    return 2;
  }
  else if (argument("--suffix")) {
    if (i >= argc - 1) return -1;
    option->suffix = argv[i + 1];
    DEBUG("file suffix: %s\n", option->suffix);
    option->number++;
    return 2;
  }
  else if (argument("-fh") || argument("--fix-hit-detect") || argument("--fix-hit-detection")) {
    if (i >= argc - 1) return -1;
    option->hit_detect = argv[i + 1];
    DEBUG("import hit detection from: %s\n", option->hit_detect);
    option->number++;
    return 2;
  }
  else if (argument("-ae") || argument("--add-event") || argument("--set-event")) {
    if (i >= argc - 2) goto add_event_syntax;
    DEBUG("setting event %s for frame %s\n", argv[i + 2], argv[i + 1]);
    if (add_to_event_set(&option->event_sets, argv[i + 1], argv[i + 2])) {
      ERROR("can't set event\n");
add_event_syntax:
      error_dbg("--add-event syntax: frame event\n");
      error_dbg("frame: frame number to add the event to\n");
      error_dbg("event: the event (eg \"itemsound=magin\")\n");
      return -2;
    }
    option->number++;
    return 3;
  }
  else if (argument("-ce") || argument("--clear-events") || argument("--clear-event") || argument("--delete-events") || argument("--delete-event")) {
    if (i >= argc - 1) return -1;
    DEBUG("clearing event list from frame %s\n", argv[i + 1]);
    if (add_to_event_set(&option->event_sets, argv[i + 1], 0)) {
      ERROR("can't set event for clearing\n");
      return -1;
    }
    option->number++;
    return 2;
  }
  else if (argument("--transplant") || argument("--transplant-bone") || argument("--transplant-bones")) {
    if (i >= argc - 4) goto transplant_syntax;
    DEBUG("setting %s transplant of %s from %s\n", argv[i + 3], argv[i + 2], argv[i + 1]);
    if (add_to_transplant_set(&option->transplant_sets, argv[i + 1], argv[i + 2], argv[i + 3], argv[i + 4])) {
      ERROR("can't add transplant\n");
transplant_syntax:
      error_dbg("--transplant syntax: file bone type\n");
      error_dbg("file: GSkin file to transplant from\n");
      error_dbg("first bone: first bone name or ID from file\n");
      error_dbg("last bone: last bone name or ID from file\n");
      error_dbg("type: normal - import as many frames as possible\n");
      error_dbg("      magic - scale framerate from file\n");
      return -2;
    }
    option->number++;
    return 5;
  }
  else if (argument("--copy-material")) {
    if (i >= argc - 3) goto copy_material_syntax;
    DEBUG("copying material %s from %s as %s\n", argv[i + 2], argv[i + 1], argv[i + 3]);
    if (add_to_copy_material_sets(&option->copy_material_sets, argv[i + 1], argv[i + 2], argv[i + 3])) {
      ERROR("can't add copy_material_set\n");
copy_material_syntax:
      error_dbg("--copy-material syntax: file material name\n");
      error_dbg("file: GSkinMesh file from which to copy material\n");
      error_dbg("material: material name or ID to copy from file\n");
      error_dbg("name: name of copied material in destination GSM\n");
      error_dbg("      use \"\" to keep the same name\n");
      return -2;
    }
    option->number++;
    return 4;
  }
  else if (argument("--rotate-bone")) {
    if (i >= argc - 5) goto rotate_bone_syntax;
    DEBUG("will rotate bone %s\n", argv[i + 1]);
    if (add_to_rotation_set(&option->rotation_sets, argv[i + 1], argv[i + 2], argv[i + 3], argv[i + 4], argv[i + 5])) {
      ERROR("can't set rotation\n");
rotate_bone_syntax:
      error_dbg("--rotate-bone syntax: bone axis angle start end\n");
      error_dbg("bone: bone name or ID to rotate\n");
      error_dbg("axis: rotation axis (X, Y or Z)\n");
      error_dbg("angle: angle to rotate (in degrees)\n");
      error_dbg("start: first frame on which to perform the rotation\n");
      error_dbg("end: last from on which to perform the rotation (or the string \"last\")\n");
      return -2;
    }
    option->number++;
    return 6;
  }
#ifdef USE_IMAGEMAGICK
  else if (argument("-uv") || argument("--uv") || argument("--generate-uv-maps")) {
    if (i >= argc - 2) goto uv_syntax;
    DEBUG("will generate UVs for group %s\n", argv[i + 1]);
    if (add_to_generate_uv_map_set(&option->generate_uv_map_sets, argv[i + 1], argv[i + 2])) {
      ERROR("can't set generate_uv_map\n");
uv_syntax:
      error_dbg("--uv syntax: group colour\n");
      error_dbg("group: triangle group ID to map\n");
      error_dbg("colour: colour to draw the maps\n");
      error_dbg("notes: colour should be in HTML format (eg #000000 for black or #ffffff for\n");
      error_dbg("white) or use \"transparent\" to generate a white-on-black mapping without\n");
      error_dbg("including the source image (import this into Photoshop and mask the background\n");
      return -2;
    }
    option->number++;
    return 3;
  }
#endif

  return 0;
}

int version() {
  SAY("gsmhacking version %s", VERSION);
#ifdef _MSC_VER
  SAY(" - Visual C++");
#endif
#ifdef __CYGWIN32__
  SAY(" - Cygwin");
#endif
#ifdef EBUG
  SAY(" DEBUG");
#endif
  SAY(" by [furrycat], %s\n", VERSION_DATE);
  return 0;
}

/* append options to option set */
void copy_options(struct option *dest, struct option *src) {
  TRACE("copying option at %p to %p\n", src, dest);
  if (src->number) dest->number++;
  if (src->convert) dest->convert = src->convert;
  if (src->add_flashlight) dest->add_flashlight = src->add_flashlight;
  if (src->make_rifle) dest->make_rifle = src->make_rifle;
  if (src->type) dest->type = src->type;
  if (src->texture_list) dest->texture_list = src->texture_list;
  if (src->showmissing) dest->showmissing = src->showmissing;
  if (src->texture_searches) gll_append((struct gll **) &dest->texture_searches, (struct gll *) src->texture_searches);
  if (src->texture_replaces) gll_append((struct gll **) &dest->texture_replaces, (struct gll *) src->texture_replaces);
  if (src->texture_sets) gll_append((struct gll **) &dest->texture_sets, (struct gll *) src->texture_sets);
  if (src->map_sets) gll_append((struct gll **) &dest->map_sets, (struct gll *) src->map_sets);
  if (src->material_sets) gll_append((struct gll **) &dest->material_sets, (struct gll *) src->material_sets);
  if (src->fd) dest->fd = src->fd;
  else if (src->outfile) dest->outfile = src->outfile;
  if (src->prefix) dest->prefix = src->prefix;
  if (src->suffix) dest->suffix = src->suffix;
  if (src->hit_detect) dest->hit_detect = src->hit_detect;
  if (src->describe) dest->describe = src->describe;
  if (src->export) dest->export = src->export;
  if (src->summary) dest->summary = src->summary;
  if (src->classify) dest->classify = src->classify;
  if (src->polycount) dest->polycount = src->polycount;
  if (src->make_stl) dest->make_stl = src->make_stl;
  if (src->optimise) dest->optimise = src->optimise;
  if (src->optimise_materials) dest->optimise_materials = src->optimise_materials;
  if (src->optimise_textures) dest->optimise_textures = src->optimise_textures;
  if (src->frame_repeat_every) dest->frame_repeat_every = src->frame_repeat_every;
  if (src->frame_repeat_count) dest->frame_repeat_count = src->frame_repeat_count;
  if (src->event_sets) gll_append((struct gll **) &dest->event_sets, (struct gll *) src->event_sets);
  if (src->transplant_sets) gll_append((struct gll **) &dest->transplant_sets, (struct gll *) src->transplant_sets);
  if (src->copy_material_sets) gll_append((struct gll **) &dest->copy_material_sets, (struct gll *) src->copy_material_sets);
  if (src->rotation_sets) gll_append((struct gll **) &dest->rotation_sets, (struct gll *) src->rotation_sets);
#ifdef USE_IMAGEMAGICK
  if (src->generate_uv_map_sets) gll_append((struct gll **) &dest->generate_uv_map_sets, (struct gll *) src->generate_uv_map_sets);
#endif
  if (! dest->number) if (src->number) dest->number++;
  TRACE("finished\n");
}

void debug_options(struct option *option) {
  if (option->number) {
    DEBUG("option->number: %u\n", option->number);
  }
  if (option->convert) {
    DEBUG("option->convert: %u\n", option->convert);
  }
  if (option->add_flashlight) {
    DEBUG("option->add_flashlight: %u\n", option->add_flashlight);
  }
  if (option->make_rifle) {
    DEBUG("option->make_rifle: %u\n", option->make_rifle);
  }
  if (option->type) {
    DEBUG("option->type: %u\n", option->type);
  }
  if (option->texture_list) {
    DEBUG("option->texture_list: %u\n", option->texture_list);
  }
  if (option->showmissing) {
    DEBUG("option->showmissing: %u\n", option->showmissing);
  }
  if (option->texture_searches) {
    DEBUG("option->texture_searches: %p next: %p\n", option->texture_searches, option->texture_searches);
  }
  if (option->texture_replaces) {
    DEBUG("option->texture_replaces: %p next: %p\n", option->texture_replaces, option->texture_replaces);
  }
  if (option->texture_sets) {
    DEBUG("option->texture_sets: %p next: %p\n", option->texture_sets, option->texture_sets->next);
  }
  if (option->map_sets) {
    DEBUG("option->map_sets: %p next: %p\n", option->map_sets, option->map_sets->next);
  }
  if (option->material_sets) {
    DEBUG("option->material_sets: %p next: %p\n", option->material_sets, option->material_sets->next);
  }
  if (option->outfile) {
    DEBUG("option->outfile: %s\n", option->outfile);
  }
  if (option->prefix) {
    DEBUG("option->prefix: %s\n", option->prefix);
  }
  if (option->suffix) {
    DEBUG("option->suffix: %s\n", option->suffix);
  }
  if (option->hit_detect) {
    DEBUG("option->hit_detect: %s\n", option->hit_detect);
  }
  if (option->describe) {
    DEBUG("option->describe: %u\n", option->describe);
  }
  if (option->export) {
    DEBUG("option->export: %u\n", option->export);
  }
  if (option->summary) {
    DEBUG("option->summary: %u\n", option->summary);
  }
  if (option->polycount) {
    DEBUG("option->polycount: %u\n", option->polycount);
  }
  if (option->classify) {
    DEBUG("option->classify: %u\n", option->classify);
  }
  if (option->optimise_materials) {
    DEBUG("option->optimise_materials: %u\n", option->optimise_materials);
  }
  if (option->optimise_textures) {
    DEBUG("option->optimise_textures: %u\n", option->optimise_textures);
  }
  if (option->optimise) {
    DEBUG("option->optimise: %u\n", option->optimise);
  }
  if (option->frame_repeat_every && option->frame_repeat_count) {
    DEBUG("option->frame_repeat_every: %u\n", option->frame_repeat_every);
    DEBUG("option->frame_repeat_count: %u\n", option->frame_repeat_count);
  }
  if (option->event_sets) {
    DEBUG("option->event_sets: %p next: %p\n", option->event_sets, option->event_sets->next);
  }
  if (option->transplant_sets) {
    DEBUG("option->transplant_sets: %p next: %p\n", option->transplant_sets, option->transplant_sets->next);
  }
  if (option->copy_material_sets) {
    DEBUG("option->copy_material_sets: %p next: %p\n", option->copy_material_sets, option->copy_material_sets->next);
  }
  if (option->rotation_sets) {
    DEBUG("option->rotation_sets: %p next: %p\n", option->rotation_sets, option->rotation_sets->next);
  }
#ifdef USE_IMAGEMAGICK
  if (option->generate_uv_map_sets) {
    DEBUG("option->generate_uv_map_sets: %p next: %p\n", option->generate_uv_map_sets, option->generate_uv_map_sets->next);
  }
#endif
}
