#include "gsmhacking.h"

struct shared_triangle_list **alloc_shared_triangle_lists(unsigned short n) {
  return (struct shared_triangle_list **) alloc_lots(n, "shared_triangle_list", sizeof(struct shared_triangle_list *), (void *(*)()) alloc_shared_triangle_list);
}

struct shared_triangle_list **alloc_more_shared_triangle_lists(unsigned short n, unsigned short num_shared_triangle_lists, struct shared_triangle_list **shared_triangle_lists) {
  return (struct shared_triangle_list **) alloc_more(n, "shared_triangle_list", sizeof(struct shared_triangle_list *), (void *(*)()) alloc_shared_triangle_list, num_shared_triangle_lists, (void **) shared_triangle_lists, sizeof(struct shared_triangle_list));
}

struct shared_triangle_list *alloc_shared_triangle_list() {
  return (struct shared_triangle_list *) alloc_one("shared_triangle_list", sizeof(struct shared_triangle_list));
}

void cleanup_shared_triangle_lists(unsigned short num_shared_triangle_lists, struct shared_triangle_list **shared_triangle_lists) {
  unsigned short i, j;
  for (i = 0; i < num_shared_triangle_lists; i++) free_structs("shared_triangle", shared_triangle_lists[i]->num_triangles, (void **) shared_triangle_lists[i]->triangles);
  free_structs("shared_triangle_list", num_shared_triangle_lists, (void **) shared_triangle_lists);
}

int find_shared_triangle_lists(struct gsm *gsm) {
  unsigned short i;
  unsigned int count;
  unsigned int pos;
  struct shared_triangle_list *shared_triangle_list;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  if (read_int(gsm->fd, &gsm->shared_triangle_list_length)) {
    ERROR("can't read shared triangle list length!\n");
    return 2;
  }
  /* number of ->shorts<- read */
  count = 0;

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think shared triangle lists start at 0x%04x\n", pos);

  /* how many triangle lists are there? */
  if (read_short(gsm->fd, &gsm->num_shared_triangle_lists)) {
    ERROR("can't read number of shared riangle lists");
    return 3;
  }
  if (! gsm->num_shared_triangle_lists) {
    SAY("%s has no shared triangle lists\n", gsm->filename);
    return 4;
  }
  DEBUG("shared triangle lists: %u\n", gsm->num_shared_triangle_lists);
  count++;

  /* allocate a big enough struct */
  gsm->shared_triangle_lists = alloc_shared_triangle_lists(gsm->num_shared_triangle_lists);
  if (! gsm->shared_triangle_lists) return 4;

  /* fill in the shared_triangle_lists */
  for (i = 0; i < gsm->num_shared_triangle_lists; i++) {
    shared_triangle_list = gsm->shared_triangle_lists[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    shared_triangle_list->offset = pos;
    if (parse_shared_triangle_list(gsm, i, shared_triangle_list, gsm->num_shared_triangle_lists, &count)) {
      ERROR("expected shared triangle lists id %u but got null\n", i);
      return 6;
    }
  }
  DEBUG("shorts read: %u (expected %u), position: 0x%04x\n", count, gsm->shared_triangle_list_length, get_file_position(gsm->fd));

  return 0;
}

int parse_shared_triangle_list(struct gsm *gsm, unsigned short num, struct shared_triangle_list *shared_triangle_list, unsigned short total, unsigned int *count) {
  unsigned short i;

  if (read_short(gsm->fd, &shared_triangle_list->num_triangles)) {
    ERROR("read failed parsing shared triangle list number of triangles\n");
    return 1;
  }
  ++*count;

  /* XXX: make shared_triangle struct in own C file? */
  shared_triangle_list->triangles = (unsigned short **) alloc_lots(shared_triangle_list->num_triangles, "shared_triangle", sizeof(unsigned short *), (void *(*)()) alloc_short);

  if (! shared_triangle_list->triangles) {
    ERROR("out of memory for list");
    return 2;
  }
  for (i = 0; i < shared_triangle_list->num_triangles; i++) {
    if (read_short(gsm->fd, shared_triangle_list->triangles[i])) {
      ERROR("read failed parsing shared triangle %u\n", i);
      return 3;
    }
    ++*count;
  }

  TRACE("triangle list %u at 0x%02x: triangles: %u\n", num, shared_triangle_list->offset, shared_triangle_list->num_triangles);

  return 0;
}

int write_shared_triangle_lists(struct gsm *gsm) {
  unsigned short i, j;
  unsigned int count;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing shared triangle list length\n");
  if (write_int(gsm->out_fd, &gsm->shared_triangle_list_length)) {
    ERROR("can't write gsm shared triangle list length");
    return 2;
  }
  count = 0;

  DEBUG("writing shared triangle lists: %u\n", gsm->num_shared_triangle_lists);
  if (write_short(gsm->out_fd, &gsm->num_shared_triangle_lists)) {
    ERROR("can't write number of shared triangle lists");
    return 3;
  }
  count++;

  for (i = 0; i < gsm->num_shared_triangle_lists; i++) {
    TRACE("writing shared triangle list %u\n", i);
    if (write_short(gsm->out_fd, &gsm->shared_triangle_lists[i]->num_triangles)) {
      ERROR("can't write shared triangle list %u number of triangles", i);
      return 4;
    }
    count++;
    for (j = 0; j < gsm->shared_triangle_lists[i]->num_triangles; j++) {
      if (write_short(gsm->out_fd, gsm->shared_triangle_lists[i]->triangles[j])) {
        ERROR("can't write shared triangle list %u triangle %u", i, j);
        return 5;
      }
      count++;
    }
  }

  if (count != gsm->shared_triangle_list_length) {
    WARN("number of shorts written %u doesn't match shared list length %u\n", count, gsm->shared_triangle_list_length);
  }
  DEBUG("shorts written: %u (expected %u)\n", count, gsm->shared_triangle_list_length);

  return 0;
}

char *get_shared_triangle_list(struct shared_triangle_list *shared_triangle_list) {
  unsigned short i;
  unsigned int len;
  char *ret, *s;

  if (! shared_triangle_list->num_triangles) return "none";

  ret = (char *) malloc(sizeof(char) * 6 * shared_triangle_list->num_triangles);
  if (! ret) {
    ERROR("out of memory for triangle list entries");
    return "unknown";
  }

  s = ret;
  TRACE("getting entries: %u\n", shared_triangle_list->num_triangles);
  for (i = 0; i < shared_triangle_list->num_triangles; i++) {
    TRACE1("writing entry number %u (triangle %u) to %p\n", i, *shared_triangle_list->triangles[i], s);
    len = snprintf(s, 6, "%u", *shared_triangle_list->triangles[i]);
    TRACE2("snprintf({s=%p}, 6, \"%u\"): %u\n", s, *shared_triangle_list->triangles[i], len);
    if (len == -1) {
      ERROR("snprintf() failed for triangle name %u\n", i);
      free(ret);
      return "unknown";
    }
    s += len;
    if (i < shared_triangle_list->num_triangles - 1) {
      TRACE1("writing comma %u to %p\n", i, s);
      len = snprintf(s, 3, ", ");
      TRACE2("snprintf({s=%p}, 3, \", \"): %u\n", s, len);
      if (len == -1) {
        ERROR("snprintf() failed for comma %u\n", i);
        free(ret);
        return "unknown";
      }
    }
    TRACE2("string ret at %p is \"%s\"\n", ret, ret);
    s += len;
//  free(name);
  }

  return ret;
}

/* return the list as { (v0,v1,v2), (v3,v4,v5), (v6,v7,v8), ... } */
char *get_shared_triangle_list_as_positions(struct gsm *gsm, unsigned short shared_triangle_list_id) {
  unsigned short i;
  unsigned int len;
  char *ret, *s;
  struct shared_triangle_list *shared_triangle_list;

  if (! gsm) {
    ERROR("no GSM!\n");
    return "BOGUS";
  }

  if (shared_triangle_list_id >= gsm->num_shared_triangle_lists) {
    ERROR("shared_triangle_list_id out of range\n");
    return "BOGUS";
  }

  shared_triangle_list = gsm->shared_triangle_lists[shared_triangle_list_id];

  if (! shared_triangle_list->num_triangles) return "none";

  ret = (char *) malloc(19 * shared_triangle_list->num_triangles);
  if (! ret) {
    ERROR("out of memory for triangle list entries");
    return "unknown";
  }

  s = ret;
  TRACE("getting entries: %u\n", shared_triangle_list->num_triangles);
  for (i = 0; i < shared_triangle_list->num_triangles; i++) {
    TRACE1("writing entry number %u (triangle %u) to %p\n", i, *shared_triangle_list->triangles[i], s);
    len = snprintf(s, 19, "%s", get_triangle_as_positions(gsm, *shared_triangle_list->triangles[i]));
    TRACE2("snprintf({s=%p}, 19, \"%s\"): %u\n", s, *shared_triangle_list->triangles[i], len);
    if (len == -1) {
      ERROR("snprintf() failed for triangle name %u\n", i);
      free(ret);
      return "unknown";
    }
    s += len;
    if (i < shared_triangle_list->num_triangles - 1) {
      TRACE1("writing comma %u to %p\n", i, s);
      len = snprintf(s, 3, ", ");
      TRACE2("snprintf({s=%p}, 3, \", \"): %u\n", s, len);
      if (len == -1) {
        ERROR("snprintf() failed for comma %u\n", i);
        free(ret);
        return "unknown";
      }
    }
    TRACE2("string ret at %p is \"%s\"\n", ret, ret);
    s += len;
//  free(name);
  }

  return ret;
}

/*
  finds the positions shared amongst the triangles of the shared list and puts 
  them into num, returns number found
*/
int get_shared_triangle_list_shared_positions(struct gsm *gsm, unsigned short shared_triangle_list_id, int **num) {
  struct shared_triangle_list *shared_triangle_list;
  struct triangle *triangle;
  int count, ret, v;
  unsigned short i;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -2;
  }

  if (shared_triangle_list_id >= gsm->num_shared_triangle_lists) {
    ERROR("shared_triangle_list_id out of range\n");
    return -3;
  }

  shared_triangle_list = gsm->shared_triangle_lists[shared_triangle_list_id];

  if (shared_triangle_list->num_triangles < 2) {
    DEBUG("called with fewer than 2 triangles!\n");
    return -1;
  }

  count = 0;
  /*
    assume the first vertex is the one shared between all triangles in the 
    list and work through each triangle, bailing out if one doesn't contain 
    this vertex
  */
  triangle = gsm->triangles[*shared_triangle_list->triangles[0]];
  if (! triangle) {
    ERROR("first triangle in list %u is null\n", shared_triangle_list_id);
    return -4;
  }

  /* test inclusion in triangle for all vertices */
  if (position_in_triangles(gsm, shared_triangle_list, triangle->vertex1, num, &count)) return -5;
  if (position_in_triangles(gsm, shared_triangle_list, triangle->vertex2, num, &count)) return -6;
  if (position_in_triangles(gsm, shared_triangle_list, triangle->vertex3, num, &count)) return -7;

  if (! count) TRACE2("no positions present in all triangles in list %u (previously thought to be impossible)\n", shared_triangle_list_id);

  return count;
}

int position_in_triangles(struct gsm *gsm, struct shared_triangle_list *shared_triangle_list, unsigned short vertex_id, int **num, int *count) {
  int ret, v, i;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -1;
  }
  if (! shared_triangle_list) {
    ERROR("no shared triangle list\n");
    return -2;
  }
  if (! num) {
    ERROR("null **num\n");
    return -3;
  }
  if (! count) {
    ERROR("null *count\n");
    return -4;
  }

  v = get_position_by_vertex(gsm, vertex_id);
  if (v < 0) {
    ERROR("umm, no position for vertex %d!?!\n", vertex_id);
    return -5;
  }
  TRACE("checking for position %d\n", v);

  for (i = 1; i < shared_triangle_list->num_triangles; i++) {
    ret = get_triangle_vertex_by_position(gsm, *shared_triangle_list->triangles[i], v);
    if (ret < 0) {
      ERROR("get_triangle_vertex_by_position() returned %d\n", ret);
      return -6;
    }
    if (! ret) {
      v = -1;
      break;
    }
  }
  if (v > -1) {
    *num[*count] = v;
    ++*count;
  }

  return 0;
}

/* XXX: first attempt at making a shared triangle list */
int make_shared_triangle_list(struct gsm *gsm) {
  unsigned short i,j, last;
  int ret;
  struct shared_triangle_list *shared_triangle_list;

  /* actually for our very first attempt we'll just print stuff to stdout :-) */
  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  INFO("creating bogus shared triangle lists\n");
  gsm->num_shared_triangle_lists = gsm->num_vertices;
  gsm->shared_triangle_lists = alloc_shared_triangle_lists(gsm->num_shared_triangle_lists);
  if (! gsm->shared_triangle_lists) {
    ERROR("out of memory for shared triangle lists\n");
    return -1;
  }

  for (i = 0; i < gsm->num_vertices; i++) {
    gsm->vertices[i]->shared_triangle_list_id = i;
    shared_triangle_list = gsm->shared_triangle_lists[i];
    shared_triangle_list->num_triangles = 1;
    shared_triangle_list->triangles = (unsigned short **) alloc_lots(shared_triangle_list->num_triangles, "shared_triangle", sizeof(unsigned short *), (void *(*)()) alloc_short);
    if (! shared_triangle_list->triangles) {
      ERROR("out of memory for list");
      return -2;
    }
    *shared_triangle_list->triangles[0] = i;
  }
  gsm->shared_triangle_list_length = -2  + gsm->num_shared_triangle_lists * 2;
  gsm->shared_triangle_list_length = (2 + gsm->num_shared_triangle_lists * 4) / 2;

  return 0;

  /* FIXME */

  INFO("attempting to generate shared triangle list by gathering triangles by position\nfor all positions in order\n");
  for (i = 0; i < gsm->num_positions; i++) {
    for (j = 0; j < gsm->num_triangles; j++) {
      ret = get_triangle_vertex_by_position(gsm, j, i);
      if (ret < 0) {
        ERROR("came unstuck finding triangles containing position %u\n", i);
        return 2;
      }
      else if (ret) {
        /* XXX :do something useful */
        DEBUG("position %u: triangle %u\n", i, j);
      }
    }
  }

  return 0;
}

int describe_shared_triangle_lists(struct gsm *gsm) {
  unsigned short i, j;
  struct shared_triangle_list *shared_triangle_list;
  int ret;
  int **num;
  char *tmp;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Shared triangle stream length: %d in %u lists\n", gsm->shared_triangle_list_length * 2 - 2, gsm->num_shared_triangle_lists);
  if (gsm_verbose) {
    num = (int **) malloc(sizeof(int *) * 3);
    if (! num) {
      ERROR("out of memory for num");
      return -1;
    }
    for (i = 0; i < 3; i++) {
      num[i] = (int *) alloc_one("int", sizeof(int));
      if (! num[i]) {
        ERROR("out of memory for num[%u]\n", i);
        return -1;
      }
    }
    for (i = 0; i < gsm->num_shared_triangle_lists; i++) {
      shared_triangle_list = gsm->shared_triangle_lists[i];
      tmp = get_shared_triangle_list(shared_triangle_list);
      ret = get_shared_triangle_list_shared_positions(gsm, i, num);
      SAY("  Shared list %02u at 0x%04x: %s", i, shared_triangle_list->offset, tmp);
      free(tmp);
      if (ret > 0) {
          if (ret == 1) SAY(" sharing position %d\n", *num[0]);
        else if (ret == 2) SAY(" sharing positions %d and %d\n", *num[0], *num[1]);
        else if (ret == 3) SAY(" sharing positions %d, %d and %d\n", *num[0], *num[1], *num[2]);
      }
      else SAY("\n");
      if (gsm_verbose > 2) {
        for (j = 0; j < shared_triangle_list->num_triangles; j++) {
          tmp = get_triangle_as_positions(gsm, *shared_triangle_list->triangles[j]);
          SAY("                            %s\n", tmp);
          free(tmp);
        }
      }
    }
    free_structs("int", 3, (void **) num);
  }

  return 0;
}
