#ifndef TRANSPLANT_H
#define TRANSPLANT_H
#include "data.h"
#include "maths.h"
#include "mapping.h"

#define TRANSPLANT_TYPE_NORMAL 0
#define TRANSPLANT_TYPE_MAGIC 1

/* transplant bones from another file */
struct transplant_set {
  struct transplant_set *next;
  char *file_arg;
  char *first_bone_arg;
  char *last_bone_arg;
  struct gsm *gsm;
  unsigned short first_bone_id;
  unsigned short last_bone_id;
  unsigned char type;
};

/* copy a material from another file (use for adding sunglasses/NVGs etc) */
struct copy_material_set {
  struct copy_material_set *next;
  char *file_arg;
  char *material_arg;
  char *name_arg;
  struct gsm *gsm;
  unsigned short material_id;
};

GSM_API struct transplant_set *alloc_transplant_set();
GSM_API void cleanup_transplant_set(struct transplant_set *);
GSM_API int add_to_transplant_set(struct transplant_set **, char *, char *, char *, char *);
GSM_API int transplant_bones(struct gsm *);
GSM_API int anim_transplant_bones(struct gsm *);
GSM_API int anim_prep_bone_transplant(struct gsm *, struct transplant_set *);
GSM_API int anim_perform_bone_transplant(struct gsm *, struct gsm *, struct transplant_set *);
GSM_API int anim_transplant_bone(struct gsm *, struct bone *, struct bone *, unsigned short);
//GSM_API void anim_transplant_bone(struct gsm *, struct gsm *, unsigned short, quaternion_t *, quaternion_t *);
GSM_API struct copy_material_set *alloc_copy_material_set();
GSM_API void cleanup_copy_material_sets(struct copy_material_set *);
GSM_API int add_to_copy_material_set(struct copy_material_set **, char *, char *, char *);
GSM_API int copy_materials(struct gsm *);
GSM_API int copy_material(struct gsm *, struct copy_material_set *);
GSM_API int copy_material_create_material(struct gsm *, struct gsm *, struct copy_material_set *);
GSM_API int copy_material_create_triangles(struct gsm *, struct gsm *, struct copy_material_set *, unsigned short *, unsigned short *);
GSM_API int copy_material_create_vertices(struct gsm *, struct gsm *, unsigned short, unsigned short, mapping_t **);
GSM_API int copy_material_create_texture_coords(struct gsm *, struct gsm *, unsigned short, mapping_t **);
GSM_API int copy_material_create_shared_triangle_lists(struct gsm *, struct gsm *, unsigned short, unsigned short);
GSM_API int copy_material_create_positions(struct gsm *, struct gsm *, unsigned short, mapping_t **);
GSM_API int copy_material_insert_positions(struct gsm *, struct gsm *, mapping_t *, unsigned short);

#endif
