#include "gsmhacking.h"

struct triangle_group **alloc_triangle_groups(unsigned short n) {
  return (struct triangle_group **) alloc_lots(n, "triangle_group", sizeof(struct triangle_group *), (void *(*)()) alloc_triangle_group);
}

struct triangle_group **alloc_more_triangle_groups(unsigned short n, unsigned short num_triangle_groups, struct triangle_group **triangle_groups) {
  return (struct triangle_group **) alloc_more(n, "triangle_group", sizeof(struct triangle_group *), (void *(*)()) alloc_triangle_group, num_triangle_groups, (void **) triangle_groups, sizeof(struct triangle_group));
}

struct triangle_group *alloc_triangle_group() {
  return (struct triangle_group *) alloc_one("triangle_group", sizeof(struct triangle_group));
}

void cleanup_triangle_groups(unsigned short num_triangle_groups, struct triangle_group **triangle_groups) {
  free_structs("triangle_group", num_triangle_groups, (void *) triangle_groups);
}

int find_triangle_groups(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct triangle_group *triangle_group;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think triangle groups start at 0x%04x\n", pos);

  /* how many triangle groups are there? */
  if (read_short(gsm->fd, &gsm->num_triangle_groups)) {
    ERROR("can't read number of triangle groups");
    return 2;
  }
  if (! gsm->num_triangle_groups) {
    SAY("%s has no triangle groups\n", gsm->filename);
    return 3;
  }
  DEBUG("triangle groups: %u\n", gsm->num_triangle_groups);

  /* allocate a big enough struct */
  gsm->triangle_groups = alloc_triangle_groups(gsm->num_triangle_groups);
  if (! gsm->triangle_groups) return 4;

  /* fill in the triangle_groups */
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    triangle_group = gsm->triangle_groups[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    triangle_group->offset = pos;
    if (parse_triangle_group(gsm, i, triangle_group, gsm->num_triangle_groups)) {
      ERROR("expected triangle groups id %u but got null\n", i);
      return 6;
    }
  }
  if (gsm->num_triangles) {
    if (gsm->num_triangles != triangle_group->first_triangle + triangle_group->num_triangles) {
      ERROR("number of triangles in triangle groups (%u) doesn't match number of triangles (%u)\n", triangle_group->first_triangle + triangle_group->num_triangles, gsm->num_triangles);
    }
  }
  else WARN("gsm->num_triangles undefined\n");

  return 0;
}

int parse_triangle_group(struct gsm *gsm, unsigned short num, struct triangle_group *triangle_group, unsigned short total) {
  if (read_short(gsm->fd, &triangle_group->material_id)) {
    ERROR("read failed parsing triangle group material number\n");
    return 1;
  }

  if (read_short(gsm->fd, &triangle_group->first_triangle)) {
    ERROR("read failed parsing triangle group first triangle\n");
    return 2;
  }

  if (read_short(gsm->fd, &triangle_group->num_triangles)) {
    ERROR("read failed parsing triangle group number of triangles\n");
    return 3;
  }

  if (read_short(gsm->fd, &triangle_group->unknown)) {
    ERROR("read failed parsing triangle group unknown\n");
    return 4;
  }
  format_unknown("triangle_group", (unsigned char *) &triangle_group->unknown, sizeof(short));

  TRACE("triangle group %u at 0x%04x: triangles: %u from %u with material %u\n", num, triangle_group->offset, triangle_group->num_triangles, triangle_group->first_triangle, triangle_group->material_id);

  return 0;
}

int write_triangle_groups(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing triangle groups: %u\n", gsm->num_triangle_groups);
  if (write_short(gsm->out_fd, &gsm->num_triangle_groups)) {
    ERROR("can't write number of triangle groups");
    return 2;
  }

  for (i = 0; i < gsm->num_triangle_groups; i++) {
    TRACE("writing triangle_group %u\n", i);
    if (write_short(gsm->out_fd, &gsm->triangle_groups[i]->material_id)) {
      ERROR("can't write triangle group %u material number", i);
      return 3;
    }
    if (write_short(gsm->out_fd, &gsm->triangle_groups[i]->first_triangle)) {
      ERROR("can't write triangle group %u first triangle", i);
      return 4;
    }
    if (write_short(gsm->out_fd, &gsm->triangle_groups[i]->num_triangles)) {
      ERROR("can't write triangle group %u number of triangles", i);
      return 5;
    }
    if (write_short(gsm->out_fd, &gsm->triangle_groups[i]->unknown)) {
      ERROR("can't write triangle group %u unknown", i);
      return 6;
    }
  }

  return 0;
}

/* how many groups are mapped to this material? */
int get_num_triangle_groups(struct gsm *gsm, unsigned short material_id) {
  unsigned short i, count;

  if (! gsm || gsm->num_materials <= material_id) {
    ERROR("no material in gsm!\n");
    return -1;
  }

  count = 0;
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    if (gsm->triangle_groups[i]->material_id == material_id) count++;
  }

  return count;
}

void zero_vertex_count/*_for_triangle_group*/(struct gsm *gsm/*, unsigned short material_id*/) {
  unsigned short i, j;

  if (! gsm/* || gsm->num_materials <= material_id*/) {
    ERROR("no material in gsm!\n");
    return;
  }

  for (i = 0; i < gsm->num_triangle_groups; i++) {
    /*if (gsm->triangle_groups[i]->material_id != material_id) continue;*/
    for (j = 0; j < gsm->triangle_groups[i]->num_triangles; j++) {
      gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex1]->count = -1;
      gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex2]->count = -1;
      gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex3]->count = -1;
    }
  }
}

int get_vertices_by_material(struct gsm *gsm, unsigned short material_id) {
  unsigned short i, j, count;

  if (! gsm || gsm->num_materials <= material_id) {
    ERROR("no material in gsm!\n");
    return -1;
  }

  /* counted stores the index relative to this "mesh" in Milkshape terms */

  /* first initialise position counters */
  zero_vertex_count/*_for_triangle_group*/(gsm/*, material_id*/);
  /* and then actually count */
  count = 0;
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    if (gsm->triangle_groups[i]->material_id != material_id) continue;
    for (j = 0; j < gsm->triangle_groups[i]->num_triangles; j++) {
      /* increase first position count */
      if (gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex1]->count < 0) gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex1]->count = count++;
      /* increase second position count */
      if (gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex2]->count < 0) gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex2]->count = count++;
      /* increase thrid position count */
      if (gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex3]->count < 0) gsm->vertices[gsm->triangles[gsm->triangle_groups[i]->first_triangle + j]->vertex3]->count = count++;
    }
  }

  return count;
}

int get_triangles_by_material(struct gsm *gsm, unsigned short material_id) {
  unsigned short i, j, count;

  if (! gsm || gsm->num_materials <= material_id) {
    ERROR("no material in gsm!\n");
    return -1;
  }

  count = 0;
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    if (gsm->triangle_groups[i]->material_id == material_id) {
      count += gsm->triangle_groups[i]->num_triangles;
    }
  }

  return count;
}

int describe_triangle_groups(struct gsm *gsm) {
  unsigned short i;
  struct triangle_group *triangle_group;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Triangle groups: %u\n", gsm->num_triangle_groups);
  if (gsm_verbose) {
    for (i = 0; i < gsm->num_triangle_groups; i++) {
      triangle_group = gsm->triangle_groups[i];
      SAY("  Triangle group %04u at 0x%04x: triangles: %u from %u to %u with %02d %s\n", i, triangle_group->offset, triangle_group->num_triangles, triangle_group->first_triangle, triangle_group->num_triangles + triangle_group->first_triangle - 1, triangle_group->material_id, get_material_name(gsm, triangle_group->material_id));
    }
  }

  return 0;
}

int get_triangle_group_by_triangle(struct gsm *gsm, unsigned short triangle_id) {
  unsigned short i;
  struct triangle_group *triangle_group;

  if (! gsm || gsm->num_triangles <= triangle_id) {
    ERROR("no triangle in gsm!\n");
    return -1;
  }

  for (i = 0; i < gsm->num_triangle_groups; i++) {
    triangle_group = gsm->triangle_groups[i];
    if (triangle_id >= triangle_group->first_triangle && triangle_id <= triangle_group->first_triangle + triangle_group->num_triangles) return (int) i;
  }

  return -1;
}
